####################################################################################################
# neuropythy/geometry/util.py
# This file defines common rotation functions that are useful with cortical mesh spheres, such as
# those produced with FreeSurfer.
# By Noah C. Benson

import numpy as np
import math, six

from ..util import (czdivide, zinv)

def normalize(u):
    '''
    normalize(u) yields a vetor with the same direction as u but unit length, or, if u has zero
    length, yields u.
    '''
    u = np.asarray(u)
    unorm = np.sqrt(np.sum(u**2, axis=0))
    z = np.isclose(unorm, 0)
    c = np.logical_not(z) / (unorm + z)
    return u * c

def vector_angle_cos(u, v):
    '''
    vector_angle_cos(u, v) yields the cosine of the angle between the two vectors u and v. If u
    or v (or both) is a (d x n) matrix of n vectors, the result will be a length n vector of the
    cosines.
    '''
    u = np.asarray(u)
    v = np.asarray(v)
    return (u * v).sum(0) / np.sqrt((u ** 2).sum(0) * (v ** 2).sum(0))

def vector_angle(u, v, direction=None):
    '''
    vector_angle(u, v) yields the angle between the two vectors u and v. The optional argument 
    direction is by default None, which specifies that the smallest possible angle between the
    vectors be reported; if the vectors u and v are 2D vectors and direction parameters True and
    False specify the clockwise or counter-clockwise directions, respectively; if the vectors are
    3D vectors, then direction may be a 3D point that is not in the plane containing u, v, and the
    origin, and it specifies around which direction (u x v or v x u) the the counter-clockwise angle
    from u to v should be reported (the cross product vector that has a positive dot product with
    the direction argument is used as the rotation axis).
    '''
    if direction is None:
        return np.arccos(vector_angle_cos(u, v))
    elif direction is True:
        return np.arctan2(v[1], v[0]) - np.arctan2(u[1], u[0])
    elif direction is False:
        return np.arctan2(u[1], u[0]) - np.arctan2(v[1], v[0])
    else:
        axis1 = normalize(u)
        axis2 = normalize(np.cross(u, v))
        if np.dot(axis2, direction) < 0:
            axis2 = -axis2
        return np.arctan2(np.dot(axis2, v), np.dot(axis1, v))

def spherical_distance(pt0, pt1):
    '''
    spherical_distance(a, b) yields the angular distance between points a and b, both of which
      should be expressed in spherical coordinates as (longitude, latitude).
    If a and/or b are (2 x n) matrices, then the calculation is performed over all columns.
    The spherical_distance function uses the Haversine formula; accordingly it may suffer from
    rounding errors in the case of nearly antipodal points.
    '''
    dtheta = pt1[0] - pt0[0]
    dphi   = pt1[1] - pt0[1]
    a = np.sin(dphi/2)**2 + np.cos(pt0[1]) * np.cos(pt1[1]) * np.sin(dtheta/2)**2
    return 2 * np.arcsin(np.sqrt(a))
    
def rotation_matrix_3D(u, th):
    """
    rotation_matrix_3D(u, t) yields a 3D numpy matrix that rotates any vector about the axis u
    t radians counter-clockwise.
    """
    # normalize the axis:
    u = normalize(u)
    # We use the Euler-Rodrigues formula;
    # see https://en.wikipedia.org/wiki/Euler-Rodrigues_formula
    a = math.cos(0.5 * th)
    s = math.sin(0.5 * th)
    (b, c, d) = -s * u
    (a2, b2, c2, d2) = (a*a, b*b, c*c, d*d)
    (bc, ad, ac, ab, bd, cd) = (b*c, a*d, a*c, a*b, b*d, c*d)
    return np.array([[a2 + b2 - c2 - d2, 2*(bc + ad),         2*(bd - ac)],
                     [2*(bc - ad),       a2 + c2 - b2 - d2,   2*(cd + ab)],
                     [2*(bd + ac),       2*(cd - ab),         a2 + d2 - b2 - c2]])

def rotation_matrix_2D(th):
    '''
    rotation_matrix_2D(th) yields a 2D numpy rotation matrix th radians counter-clockwise about the
    origin.
    '''
    s = np.sin(th)
    c = np.cos(th)
    return np.array([[c, -s], [s, c]])

# construct a rotation matrix of vector u to vector b around center
def alignment_matrix_3D(u, v):
    '''
    alignment_matrix_3D(u, v) yields a 3x3 numpy array that rotates the vector u to the vector v
    around the origin.
    '''
    # normalize both vectors:
    u = normalize(u)
    v = normalize(v)
    # get the cross product of u cross v
    w = np.cross(u, v)
    # the angle we need to rotate
    th = vector_angle(u, v)
    # we rotate around this vector by the angle between them
    return rotation_matrix_3D(w, th)

def alignment_matrix_2D(u, v):
    '''
    alignment_matrix_2D(u, v) yields a 2x2 numpy array that rotates the vector u to the vector v
    around the origin.
    '''
    return rotation_matrix_2D(vector_angle(u, v, direction=True))

def point_on_line(ab, c, atol=1e-8):
    '''
    point_on_line((a,b), c) yields True if point x is on line (a,b) and False otherwise.
    '''
    (a,b) = ab
    abc = [np.asarray(u) for u in (a,b,c)]
    if any(len(u.shape) == 2 for u in abc): (a,b,c) = [np.reshape(u,(len(u),-1)) for u in abc]
    else:                                   (a,b,c) = abc
    vca = a - c
    vcb = b - c
    uba = czdivide(vba, np.sqrt(np.sum(vba**2, axis=0)))
    uca = czdivide(vca, np.sqrt(np.sum(vca**2, axis=0)))
    return (np.isclose(np.sqrt(np.sum(vca**2, axis=0)), 0, atol=atol) |
            np.isclose(np.sqrt(np.sum(vcb**2, axis=0)), 0, atol=atol) |
            np.isclose(np.abs(np.sum(uba*uca, axis=0)), 1, atol=atol))

def point_on_segment(ac, b, atol=1e-8):
    '''
    point_on_segment((a,b), c) yields True if point x is on segment (a,b) and False otherwise. Note
    that this differs from point_in_segment in that a point that if c is equal to a or b it is
    considered 'on' but not 'in' the segment.
    The option atol can be given and is used only to test for difference from 0; by default it is
    1e-8.
    '''
    (a,c) = ac
    abc = [np.asarray(u) for u in (a,b,c)]
    if any(len(u.shape) > 1 for u in abc): (a,b,c) = [np.reshape(u,(len(u),-1)) for u in abc]
    else:                                  (a,b,c) = abc
    vab = b - a
    vbc = c - b
    vac = c - a
    dab = np.sqrt(np.sum(vab**2, axis=0))
    dbc = np.sqrt(np.sum(vbc**2, axis=0))
    dac = np.sqrt(np.sum(vac**2, axis=0))
    return np.isclose(dab + dbc - dac, 0, atol=atol)
def point_in_segment(ac, b, atol=1e-8):
    '''
    point_in_segment((a,b), c) yields True if point x is in segment (a,b) and False otherwise. Note
    that this differs from point_on_segment in that a point that if c is equal to a or b it is
    considered 'on' but not 'in' the segment.
    The option atol can be given and is used only to test for difference from 0; by default it is
    1e-8.
    '''
    (a,c) = ac
    abc = [np.asarray(u) for u in (a,b,c)]
    if any(len(u.shape) > 1 for u in abc): (a,b,c) = [np.reshape(u,(len(u),-1)) for u in abc]
    else:                                  (a,b,c) = abc
    vab = b - a
    vbc = c - b
    vac = c - a
    dab = np.sqrt(np.sum(vab**2, axis=0))
    dbc = np.sqrt(np.sum(vbc**2, axis=0))
    dac = np.sqrt(np.sum(vac**2, axis=0))
    return (np.isclose(dab + dbc - dac, 0, atol=atol) &
            ~np.isclose(dac - dab, 0, atol=atol) &
            ~np.isclose(dac - dbc, 0, atol=atol))

def lines_colinear(ab, cd, atol=1e-8):
    '''
    liness_colinear((a, b), (c, d)) yields True if the lines containing points (a,b) and points (c,d)
    are colinear and false otherwise. All of a, b, c, and d must be (x,y) coordinates or 2xN (x,y)
    coordinate matrices, or (x,y,z) or 3xN matrices.
    '''
    # simple check: a and b must be on (c,d)
    return point_on_line(ab, cd[0], atol=atol) & point_on_line(ab, cd[1], atol=atol)

def segments_colinear(ab, cd, atol=1e-8):
    '''
    segments_colinear_2D((a, b), (c, d)) yields True if either a or b is on the line segment (c,d) or
    if c or d is on the line segment (a,b) and the lines are colinear; otherwise yields False. All
    of a, b, c, and d must be (x,y) coordinates or 2xN (x,y) coordinate matrices, or (x,y,z) or 3xN
    matrices.
    '''
    (a,b) = ab
    (c,d) = cd
    ss = [point_on_segment(ab, c, atol=atol), point_on_segment(ab, d, atol=atol),
          point_on_segment(cd, a, atol=atol), point_on_segment(cd, b, atol=atol)]
    return np.sum(ss, axis=0) > 1

def points_close(a,b, atol=1e-8):
    '''
    points_close(a,b) yields True if points a and b are close to each other and False otherwise.
    '''
    (a,b) = [np.asarray(u) for u in (a,b)]
    if len(a.shape) == 2 or len(b.shape) == 2: (a,b) = [np.reshape(u,(len(u),-1)) for u in (a,b)]
    return np.isclose(np.sqrt(np.sum((a - b)**2, axis=0)), 0, atol=atol)

def segments_overlapping(ab, cd, atol=1e-8):
    '''
    segments_overlapping((a, b), (c, d)) yields True if the line segments (a,b) and (c,d) are both
    colinear and have a non-finite overlap. If (a,b) and (c,d) touch at a single point, they are not
    considered overlapping.
    '''
    (a,b) = ab
    (c,d) = cd
    ss = [point_in_segment(ab, c, atol=atol), point_in_segment(ab, d, atol=atol),
          point_in_segment(cd, a, atol=atol), point_in_segment(cd, b, atol=atol)]
    return (~(points_close(a,b) | points_close(c,d, atol=atol)) & 
            ((np.sum(ss, axis=0) > 1) |
             (points_close(a,c, atol=atol) & points_close(b,d, atol=atol)) |
             (points_close(a,d, atol=atol) & points_close(b,c, atol=atol))))

def line_intersection_2D(abarg, cdarg, atol=1e-8):
    '''
    line_intersection((a, b), (c, d)) yields the intersection point between the lines that pass
    through the given pairs of points. If any lines are parallel, (numpy.nan, numpy.nan) is
    returned; note that a, b, c, and d can all be 2 x n matrices of x and y coordinate row-vectors.
    '''
    ((x1,y1),(x2,y2)) = abarg
    ((x3,y3),(x4,y4)) = cdarg
    dx12 = (x1 - x2)
    dx34 = (x3 - x4)
    dy12 = (y1 - y2)
    dy34 = (y3 - y4)
    denom = dx12*dy34 - dy12*dx34
    unit = np.isclose(denom, 0, atol=atol)
    if unit is True: return (np.nan, np.nan)
    denom = unit + denom
    q12 = (x1*y2 - y1*x2) / denom
    q34 = (x3*y4 - y3*x4) / denom
    xi = q12*dx34 - q34*dx12
    yi = q12*dy34 - q34*dy12
    if   unit is False: return (xi, yi)
    elif unit is True:  return (np.nan, np.nan)
    else:
        xi = np.asarray(xi)
        yi = np.asarray(yi)
        xi[unit] = np.nan
        yi[unit] = np.nan
        return (xi, yi)

def segment_intersection_2D(p12arg, p34arg, atol=1e-8):
    '''
    segment_intersection((a, b), (c, d)) yields the intersection point between the line segments
    that pass from point a to point b and from point c to point d. If there is no intersection
    point, then (numpy.nan, numpy.nan) is returned.
    '''
    (p1,p2) = p12arg
    (p3,p4) = p34arg
    pi = np.asarray(line_intersection_2D(p12arg, p34arg, atol=atol))
    p1 = np.asarray(p1)
    p2 = np.asarray(p2)
    p3 = np.asarray(p3)
    p4 = np.asarray(p4)
    u12 = p2 - p1
    u34 = p4 - p3
    cfn = lambda px,iis: (px if iis is None or len(px.shape) == 1 or px.shape[1] == len(iis) else
                          px[:,iis])
    dfn = lambda a,b:     a[0]*b[0] + a[1]*b[1]
    sfn = lambda a,b:     ((a-b)                 if len(a.shape) == len(b.shape) else
                           (np.transpose([a])-b) if len(a.shape) <  len(b.shape) else
                           (a - np.transpose([b])))
    fn  = lambda px,iis:  (1 - ((dfn(cfn(u12,iis), sfn(         px, cfn(p1,iis))) > 0) *
                                (dfn(cfn(u34,iis), sfn(         px, cfn(p3,iis))) > 0) *
                                (dfn(cfn(u12,iis), sfn(cfn(p2,iis),          px)) > 0) *
                                (dfn(cfn(u34,iis), sfn(cfn(p4,iis),          px)) > 0)))
    if len(pi.shape) == 1:
        if not np.isfinite(pi[0]): return (np.nan, np.nan)
        bad = fn(pi, None)
        return (np.nan, np.nan) if bad else pi
    else:
        nonpar = np.where(np.isfinite(pi[0]))[0]
        bad = fn(cfn(pi, nonpar), nonpar)
        (xi,yi) = pi
        bad = nonpar[np.where(bad)[0]]
        xi[bad] = np.nan
        yi[bad] = np.nan
        return (xi,yi)

def lines_touch_2D(ab, cd, atol=1e-8):
    '''
    lines_touch_2D((a,b), (c,d)) is equivalent to lines_colinear((a,b), (c,d)) |
    numpy.isfinite(line_intersection_2D((a,b), (c,d))[0])
    '''
    return (lines_colinear(ab, cd, atol=atol) |
            np.isfinite(line_intersection_2D(ab, cd, atol=atol)[0]))

def segments_touch_2D(ab, cd, atol=1e-8):
    '''
    segmentss_touch_2D((a,b), (c,d)) is equivalent to segments_colinear((a,b), (c,d)) |
    numpy.isfinite(segment_intersection_2D((a,b), (c,d))[0])
    '''
    return (segments_colinear(ab, cd, atol=atol) |
            np.isfinite(segment_intersection_2D(ab, cd, atol=atol)[0]))

def line_segment_intersection_2D(p12arg, p34arg, atol=1e-8):
    '''
    line_segment_intersection((a, b), (c, d)) yields the intersection point between the line
    passing through points a and b and the line segment that passes from point c to point d. If
    there is no intersection point, then (numpy.nan, numpy.nan) is returned.
    '''
    (p1,p2) = p12arg
    (p3,p4) = p34arg
    pi = np.asarray(line_intersection_2D(p12arg, p34arg, atol=atol))
    p3 = np.asarray(p3)
    u34 = p4 - p3
    cfn = lambda px,iis: (px if iis is None or len(px.shape) == 1 or px.shape[1] == len(iis) else
                          px[:,iis])
    dfn = lambda a,b:     a[0]*b[0] + a[1]*b[1]
    sfn = lambda a,b:     ((a-b)                 if len(a.shape) == len(b.shape) else
                           (np.transpose([a])-b) if len(a.shape) <  len(b.shape) else
                           (a - np.transpose([b])))
    fn  = lambda px,iis:  (1 - ((dfn(cfn(u34,iis), sfn(         px, cfn(p3,iis))) > 0) *
                                (dfn(cfn(u34,iis), sfn(cfn(p4,iis),          px)) > 0)))
    if len(pi.shape) == 1:
        if not np.isfinite(pi[0]): return (np.nan, np.nan)
        bad = fn(pi, None)
        return (np.nan, np.nan) if bad else pi
    else:
        nonpar = np.where(np.isfinite(pi[0]))[0]
        bad = fn(cfn(pi, nonpar), nonpar)
        (xi,yi) = pi
        bad = nonpar[np.where(bad)[0]]
        xi[bad] = np.nan
        yi[bad] = np.nan
        return (xi,yi)

def triangle_area(a,b,c):
    '''
    triangle_area(a, b, c) yields the area of the triangle whose vertices are given by the points a,
    b, and c.
    '''
    (a,b,c) = [np.asarray(x) for x in (a,b,c)]
    sides = np.sqrt(np.sum([(p1 - p2)**2 for (p1,p2) in zip([b,c,a],[c,a,b])], axis=1))
    s = 0.5 * np.sum(sides, axis=0)
    s = np.clip(s * np.prod(s - sides, axis=0), 0.0, None)
    return np.sqrt(s)

def triangle_normal(a,b,c):
    '''
    triangle_normal(a, b, c) yields the normal vector of the triangle whose vertices are given by
      the points a, b, and c. If the points are 2D points, then 3D normal vectors are still yielded,
      that are always (0,0,1) or (0,0,-1). This function auto-threads over matrices, in which case
      they must be in equivalent orientations, and the result is returned in whatever orientation
      they are given in. In some cases, the intended orientation of the matrices is ambiguous (e.g.,
      if a, b, and c are 2 x 3 matrices), in which case the matrix is always assumed to be given in
      (dims x vertices) orientation.
    '''
    (a,b,c) = [np.asarray(x) for x in (a,b,c)]
    if len(a.shape) == 1 and len(b.shape) == 1 and len(c.shape) == 1:
        return triangle_normal(*[np.transpose([x]) for x in (a,b,c)])[:,0]
    (a,b,c) = [np.transpose([x]) if len(x.shape) == 1 else x for x in (a,b,c)]
    # find a required number of dimensions, if possible
    if a.shape[0] in (2,3):
        dims = a.shape[0]
        tx = True
    else:
        dims = a.shape[1]
        (a,b,c) = [x.T for x in (a,b,c)]
        tx = False
    n = (a.shape[1] if a.shape[1] != 1 else b.shape[1] if b.shape[1] != 1 else
         c.shape[1] if c.shape[1] != 1 else 1)
    if dims == 2:
        (a,b,c) = [np.vstack((x, np.zeros((1,n)))) for x in (a,b,c)]
    ab = normalize(b - a)
    ac = normalize(c - a)
    res = np.cross(ab, ac, axisa=0, axisb=0)
    return res.T if tx else res

def cartesian_to_barycentric_3D(tri, xy):
    '''
    cartesian_to_barycentric_3D(tri,xy) is identical to cartesian_to_barycentric_2D(tri,xy) except
    it works on 3D data. Note that if tri is a 3 x 3 x n, a 3 x n x 3 or an n x 3 x 3 matrix, the
    first dimension must always be the triangle vertices and the second 3-sized dimension must be
    the (x,y,z) coordinates.
    '''
    xy = np.asarray(xy)
    tri = np.asarray(tri)
    if len(xy.shape) == 1:
        return cartesian_to_barycentric_3D(np.transpose(np.asarray([tri]), (1,2,0)),
                                           np.asarray([xy]).T)[:,0]
    xy = xy if xy.shape[0] == 3 else xy.T
    if tri.shape[0] == 3:
        tri = tri if tri.shape[1] == 3 else np.transpose(tri, (0,2,1))
    elif tri.shape[1] == 3:
        tri = tri.T if tri.shape[0] == 3 else np.transpose(tri, (1,2,0))
    elif tri.shape[2] == 3:
        tri = np.transpose(tri, (2,1,0) if tri.shape[1] == 3 else (2,0,1))
    if tri.shape[0] != 3 or tri.shape[1] != 3:
        raise ValueError('Triangle array did not have dimensions of sizes 3 and 3')
    if xy.shape[0] != 3:
        raise ValueError('coordinate matrix did not have a dimension of size 3')
    if tri.shape[2] != xy.shape[1]:
        raise ValueError('number of triangles and coordinates must match')
    # The algorithm here is borrowed from this stack-exchange post:
    # http://gamedev.stackexchange.com/questions/23743
    # in which it is attributed to Christer Ericson's book Real-Time Collision Detection.
    v0 = tri[1] - tri[0]
    v1 = tri[2] - tri[0]
    v2 = xy - tri[0]
    d00 = np.sum(v0 * v0, axis=0)
    d01 = np.sum(v0 * v1, axis=0)
    d11 = np.sum(v1 * v1, axis=0)
    d20 = np.sum(v2 * v0, axis=0)
    d21 = np.sum(v2 * v1, axis=0)
    den = d00*d11 - d01*d01
    zero = np.isclose(den, 0)
    unit = 1 - zero
    den += zero
    l2 = unit * (d11 * d20 - d01 * d21) / den
    l3 = unit * (d00 * d21 - d01 * d20) / den
    return np.asarray([1.0 - l2 - l3, l2])
    
def cartesian_to_barycentric_2D(tri, xy, atol=1e-8):
    '''
    cartesian_to_barycentric_2D(tri, xy) yields a (2 x n) barycentric coordinate matrix 
    (or just a tuple if xy is a single (x, y) coordinate) of the first two barycentric coordinates
    for the triangle coordinate array tri. The array tri should be 3 (vertices) x 2 (coordinates) x
    n (triangles) unless xy is a tuple, in which case it should be a (3 x 2) matrix.
    '''
    xy = np.asarray(xy)
    tri = np.asarray(tri)
    if len(xy.shape) == 1:
        return cartesian_to_barycentric_2D(np.transpose(np.asarray([tri]), (1,2,0)),
                                           np.asarray([xy]).T)[:,0]
    xy = xy if xy.shape[0] == 2 else xy.T
    if tri.shape[0] == 3:
        tri = tri if tri.shape[1] == 2 else np.transpose(tri, (0,2,1))
    elif tri.shape[1] == 3:
        tri = tri.T if tri.shape[0] == 2 else np.transpose(tri, (1,2,0))
    elif tri.shape[2] == 3:
        tri = np.transpose(tri, (2,1,0) if tri.shape[1] == 2 else (2,0,1))
    if tri.shape[0] != 3 or tri.shape[1] != 2:
        raise ValueError('Triangle array did not have dimensions of sizes 3 and 2')
    if xy.shape[0] != 2:
        raise ValueError('coordinate matrix did not have a dimension of size 2')
    if tri.shape[2] != xy.shape[1]:
        raise ValueError('number of triangles and coordinates must match')
    # Okay, everything's the right shape...
    (x,y) = xy
    ((x1,y1), (x2,y2), (x3,y3)) = tri
    x_x3  = x  - x3
    x1_x3 = x1 - x3
    x3_x2 = x3 - x2
    y_y3  = y  - y3
    y1_y3 = y1 - y3
    y2_y3 = y2 - y3
    num1 = (y2_y3*x_x3  + x3_x2*y_y3)
    num2 = (-y1_y3*x_x3 + x1_x3*y_y3)
    den  = (y2_y3*x1_x3 + x3_x2*y1_y3)
    zero = np.isclose(den, 0, atol=atol)
    den += zero
    unit = 1 - zero
    l1 = unit * num1 / den
    l2 = unit * num2 / den
    return np.asarray((l1, l2))

def barycentric_to_cartesian(tri, bc):
    '''
    barycentric_to_cartesian(tri, bc) yields the d x n coordinate matrix of the given barycentric
    coordinate matrix (also d x n) bc interpolated in the n triangles given in the array tri. See
    also cartesian_to_barycentric. If tri and bc represent one triangle and coordinate, then just
    the coordinate and not a matrix is returned. The value d, dimensions, must be 2 or 3.
    '''
    bc = np.asarray(bc)
    tri = np.asarray(tri)
    if len(bc.shape) == 1:
        return barycentric_to_cartesian(np.transpose(np.asarray([tri]), (1,2,0)),
                                        np.asarray([bc]).T)[:,0]
    bc = bc if bc.shape[0] == 2 else bc.T
    if bc.shape[0] != 2: raise ValueError('barycentric matrix did not have a dimension of size 2')
    n = bc.shape[1]
    # we know how many bc's there are now; lets reorient tri to match with the last dimension as n
    if len(tri.shape) == 2:
        tri = np.transpose([tri for _ in range(n)], (1,2,0))
    # the possible orientations of tri:
    if tri.shape[0] == 3:
        if tri.shape[1] in [2,3] and tri.shape[2] == n:
            pass # default orientation
        elif tri.shape[1] == n and tri.shape[2] in [2,3]:
            tri = np.transpose(tri, (0,2,1))
        else: raise ValueError('could not deduce triangle dimensions')
    elif tri.shape[1] == 3:
        if tri.shape[0] in [2,3] and tri.shape[2] == n:
            tri = np.transpose(tri, (1,0,2))
        elif tri.shape[0] == n and tri.shape[2] in [2,3]:
            tri = np.transpose(tri, (1,2,0))
        else: raise ValueError('could not deduce triangle dimensions')
    elif tri.shape[2] == 3:
        if tri.shape[0] in [2,3] and tri.shape[1] == n:
            tri = np.transpose(tri, (2,0,1))
        elif tri.shape[0] == n and tri.shape[1] in [2,3]:
            tri = np.transpose(tri, (2,1,0))
        else: raise ValueError('could not deduce triangle dimensions')
    else: raise ValueError('At least one dimension of triangles must be 3')
    if tri.shape[0] != 3 or (tri.shape[1] not in [2,3]):
        raise ValueError('Triangle array did not have dimensions of sizes 3 and (2 or 3)')
    if tri.shape[2] != n:
        raise ValueError('number of triangles and coordinates must match')
    (l1,l2) = bc
    (p1, p2, p3) = tri
    l3 = (1 - l1 - l2)
    return np.asarray([x1*l1 + x2*l2 + x3*l3 for (x1,x2,x3) in zip(p1, p2, p3)])
    
def triangle_address(fx, pt):
    '''
    triangle_address(FX, P) yields an address coordinate (t,r) for the point P in the triangle
    defined by the (3 x d)-sized coordinate matrix FX, in which each row of the matrix is the
    d-dimensional vector representing the respective triangle vertx for triangle [A,B,C]. The
    resulting coordinates (t,r) (0 <= t <= 1, 0 <= r <= 1) address the point P such that, if t gives
    the fraction of the angle from vector AB to vector AC that is made by the angle between vectors
    AB and AP, and r gives the fraction ||AP||/||AR|| where R is the point of intersection between
    lines AP and BC. If P is a (d x n)-sized matrix of points, then a (2 x n) matrix of addresses
    is returned.
    '''
    fx = np.asarray(fx)
    pt = np.asarray(pt)
    # The triangle vectors...
    ab = fx[1] - fx[0]
    ac = fx[2] - fx[0]
    bc = fx[2] - fx[1]
    ap = np.asarray([pt_i - a_i for (pt_i, a_i) in zip(pt, fx[0])])
    # get the unnormalized distance...
    r = np.sqrt((ap ** 2).sum(0))
    # now we can find the angle...
    unit = 1 - r.astype(bool)
    t0 = vector_angle(ab, ac)
    t = vector_angle(ap + [ab_i * unit for ab_i in ab], ab)
    sint = np.sin(t)
    sindt = np.sin(t0 - t)
    # finding r0 is tricker--we use this fancy formula based on the law of sines
    q0 = np.sqrt((bc ** 2).sum(0))          # B->C distance
    beta = vector_angle(-ab, bc)            # Angle at B
    sinGamma = np.sin(math.pi - beta - t0)
    sinBeta  = np.sin(beta)
    r0 = q0 * sinBeta * sinGamma / (sinBeta * sindt + sinGamma * sint)
    return np.asarray([t/t0, r/r0])

def triangle_unaddress(fx, tr):
    '''
    triangle_unaddress(FX, tr) yields the point P, inside the reference triangle given by the 
    (3 x d)-sized coordinate matrix FX, that is addressed by the address coordinate tr, which may
    either be a 2-d vector or a (2 x n)-sized matrix.
    '''
    fx = np.asarray(fx)
    tr = np.asarray(tr)
    # the triangle vectors...
    ab = fx[1] - fx[0]
    ac = fx[2] - fx[0]
    bc = fx[2] - fx[1]
    return np.asarray([ax + tr[1]*(abx + tr[0]*bcx) for (ax, bcx, abx) in zip(fx[0], bc, ab)])

def point_in_triangle(tri, pt, atol=1e-13):
    tri = np.asarray(tri)
    pt  = np.asarray(pt)
    if len(tri.shape) == 2 and len(pt.shape) == 1:
        if len(pt) == 2:
            tol = atol
            v0 = tri[2] - tri[0]
            v1 = tri[1] - tri[0]
            v2 = pt - tri[0]
            d00 = np.dot(v0, v0)
            d01 = np.dot(v0, v1)
            d02 = np.dot(v0, v2)
            d11 = np.dot(v1, v1)
            d12 = np.dot(v1, v2)
            invDenom = (d00*d11 - d01*d01)
            if np.isclose(invDenom, 0, atol=atol): return False
            s = (d11*d02 - d01*d12) / invDenom
            if (s + tol) < 0 or (s - tol) >= 1: return False
            t = (d00*d12 - d01*d02) / invDenom
            #return False if (t + tol) < 0 or (s + t - tol) > 1 else True
            s = s + t
            tp = np.isclose(t, 0, atol=tol) or t >= 0
            tn = np.isclose(s - 1, 0, atol=tol) or s <= 1
            return tn and tp
        else:
            dp1 = np.dot(pt - tri[0], np.cross(tri[0], tri[1] - tri[0]))
            dp2 = np.dot(pt - tri[1], np.cross(tri[1], tri[2] - tri[1]))
            db3 = np.dot(pt - tri[2], np.cross(tri[2], tri[0] - tri[2]))
            return ((dp1 > 0 or np.isclose(dp1, 0, atol=atol)) and
                    (dp2 > 0 or np.isclose(dp2, 0, atol=atol)) and
                    (dp3 > 0 or np.isclose(dp3, 0, atol=atol)))
    elif len(tri.shape) == 3 and len(pt.shape) == 2:
        if len(pt) != len(tri):
            raise ValueError('the number of triangles and points must be equal')
        if pt.shape[1] == 2:
            tol = atol
            v0 = tri[:,2] - tri[:,0]
            v1 = tri[:,1] - tri[:,0]
            v2 = pt - tri[:,0]
            d00 = np.sum(v0*v0, axis=1)
            d01 = np.sum(v0*v1, axis=1)
            d02 = np.sum(v0*v2, axis=1)
            d11 = np.sum(v1*v1, axis=1)
            d12 = np.sum(v1*v2, axis=1)
            invDenom = (d00*d11 - d01*d01)
            zeros = np.isclose(invDenom, 0)
            invDenom[zeros] = 1.0
            s = (d11*d02 - d01*d12) / invDenom
            t = (d00*d12 - d01*d02) / invDenom
            #return ~((s + tol < 0) | (s - tol >= 1) | (t + tol < 0) | (s + t - tol > 1) | zeros)
            #return (~(s + tol < 0) & ~(s - tol >= 1) & ~(t + tol < 0) & ~(s + t - tol > 1) & ~zeros)
            #return ((s + tol >= 0) & (s - tol < 1) & (t + tol >= 0) & (s + t - tol <= 1) & ~zeros)
            st = s + t
            q1 = np.isclose(s, 0, atol=atol) | (s > 0)
            q2 = ~np.isclose(s - 1, 0, atol=atol) & (s < 1)
            q3 = np.isclose(t, 0, atol=atol) | (t > 0)
            q4 = np.isclose(st - 1, 0, atol=atol) | (st < 1)
            return q1 * q2 & q3 & q4 & ~zeros
        else:
            x0 = np.sum((pt - tri[:,0]) * np.cross(tri[:,0], tri[:,1] - tri[:,0], axis=1), axis=1)
            x1 = np.sum((pt - tri[:,1]) * np.cross(tri[:,1], tri[:,2] - tri[:,1], axis=1), axis=1)
            x2 = np.sum((pt - tri[:,2]) * np.cross(tri[:,2], tri[:,0] - tri[:,2], axis=1), axis=1)
            return (((x0 > 0) | np.isclose(x0, 0, atol=atol)) &
                    ((x1 > 0) | np.isclose(x1, 0, atol=atol)) &
                    ((x2 > 0) | np.isclose(x2, 0, atol=atol)))
    elif len(tri.shape) == 3 and len(pt.shape) == 1:
        return point_in_triangle(tri, np.asarray([pt for _ in tri]))
    elif len(tri.shape) == 2 and len(pt.shape) == 1:
        return point_in_triangle(np.asarray([tri for _ in pt]), pt)
    else:
        raise ValueError('triangles and pts do not have parallel shapes')

def det4D(m):
    '''
    det4D(array) yields the determinate of the given matrix array, which may have more than 2
      dimensions, in which case the later dimensions are multiplied and added point-wise.
    '''
    # I just solved this in Mathematica, copy-pasted, and replaced the string '] m' with ']*m':
    # Mathematica code: Det@Table[m[i][j], {i, 0, 3}, {j, 0, 3}]
    return (m[0][3]*m[1][2]*m[2][1]*m[3][0] - m[0][2]*m[1][3]*m[2][1]*m[3][0] -
            m[0][3]*m[1][1]*m[2][2]*m[3][0] + m[0][1]*m[1][3]*m[2][2]*m[3][0] +
            m[0][2]*m[1][1]*m[2][3]*m[3][0] - m[0][1]*m[1][2]*m[2][3]*m[3][0] -
            m[0][3]*m[1][2]*m[2][0]*m[3][1] + m[0][2]*m[1][3]*m[2][0]*m[3][1] +
            m[0][3]*m[1][0]*m[2][2]*m[3][1] - m[0][0]*m[1][3]*m[2][2]*m[3][1] -
            m[0][2]*m[1][0]*m[2][3]*m[3][1] + m[0][0]*m[1][2]*m[2][3]*m[3][1] +
            m[0][3]*m[1][1]*m[2][0]*m[3][2] - m[0][1]*m[1][3]*m[2][0]*m[3][2] -
            m[0][3]*m[1][0]*m[2][1]*m[3][2] + m[0][0]*m[1][3]*m[2][1]*m[3][2] +
            m[0][1]*m[1][0]*m[2][3]*m[3][2] - m[0][0]*m[1][1]*m[2][3]*m[3][2] -
            m[0][2]*m[1][1]*m[2][0]*m[3][3] + m[0][1]*m[1][2]*m[2][0]*m[3][3] +
            m[0][2]*m[1][0]*m[2][1]*m[3][3] - m[0][0]*m[1][2]*m[2][1]*m[3][3] -
            m[0][1]*m[1][0]*m[2][2]*m[3][3] + m[0][0]*m[1][1]*m[2][2]*m[3][3])
def det_4x3(a,b,c,d):
    '''
    det_4x3(a,b,c,d) yields the determinate of the matrix formed the given rows, which may have
      more than 1 dimension, in which case the later dimensions are multiplied and added point-wise.
      The point's must be 3D points; the matrix is given a fourth column of 1s and the resulting
      determinant is of this matrix. 
    '''
    # I just solved this in Mathematica, copy-pasted, and replaced the string '] m' with ']*m':
    # Mathematica code: Det@Table[If[j == 3, 1, i[j]], {i, {a, b, c, d}}, {j, 0, 3}]
    return (a[1]*b[2]*c[0] + a[2]*b[0]*c[1] - a[2]*b[1]*c[0] - a[0]*b[2]*c[1] -
            a[1]*b[0]*c[2] + a[0]*b[1]*c[2] + a[2]*b[1]*d[0] - a[1]*b[2]*d[0] -
            a[2]*c[1]*d[0] + b[2]*c[1]*d[0] + a[1]*c[2]*d[0] - b[1]*c[2]*d[0] -
            a[2]*b[0]*d[1] + a[0]*b[2]*d[1] + a[2]*c[0]*d[1] - b[2]*c[0]*d[1] -
            a[0]*c[2]*d[1] + b[0]*c[2]*d[1] + a[1]*b[0]*d[2] - a[0]*b[1]*d[2] -
            a[1]*c[0]*d[2] + b[1]*c[0]*d[2] + a[0]*c[1]*d[2] - b[0]*c[1]*d[2])
    
def tetrahedral_barycentric_coordinates(tetra, pt):
    '''
    tetrahedral_barycentric_coordinates(tetrahedron, point) yields a list of weights for each vertex
      in the given tetrahedron in the same order as the vertices given. If all weights are 0, then
      the point is not inside the tetrahedron.
    '''
    # I found a description of this algorithm here (Nov. 2017):
    # http://steve.hollasch.net/cgindex/geometry/ptintet.html
    tetra = np.asarray(tetra)
    pt = np.asarray(pt)
    if tetra.shape[0] != 4:
        if tetra.shape[1] == 4:
            if tetra.shape[0] == 3:
                tetra = np.transpose(tetra, (1,0) if len(tetra.shape) == 2 else (1,0,2))
            else:
                tetra = np.transpose(tetra, (1,2,0))
        elif tetra.shape[1] == 3:
            tetra = np.transpose(tetra, (2,1,0))
        else:
            tetra = np.transpose(tetra, (2,0,1))
    elif tetra.shape[1] != 3:
        tetra = np.transpose(tetra, (0,2,1))
    if pt.shape[0] != 3: pt = pt.T
    # Okay, calculate the determinants...
    d_ = det_4x3(tetra[0], tetra[1], tetra[2], tetra[3])
    d0 = det_4x3(pt,       tetra[1], tetra[2], tetra[3])
    d1 = det_4x3(tetra[0], pt,       tetra[2], tetra[3])
    d2 = det_4x3(tetra[0], tetra[1], pt,       tetra[3])
    d3 = det_4x3(tetra[0], tetra[1], tetra[2], pt)
    s_ = np.sign(d_)
    z_ = np.logical_or(np.any([s_ * si == -1 for si in np.sign([d0,d1,d2,d3])], axis=0),
                       np.isclose(d_,0))
    x_ = np.logical_not(z_)
    d_inv = x_ / (x_ * d_ + z_)
    return np.asarray([d_inv * dq for dq in (d0,d1,d2,d3)])

def point_in_tetrahedron(tetra, pt):
    '''
    point_in_tetrahedron(tetrahedron, point) yields True if the given point is in the given 
      tetrahedron. If either tetrahedron or point (or both) are lists of shapes/points, then this
      calculation is automatically threaded over all the given arguments.
    '''
    bcs = tetrahedral_barycentric_coordinates(tetra, pt)
    return np.logical_not(np.all(np.isclose(bcs, 0), axis=0))

def prism_barycentric_coordinates(tri1, tri2, pt):
    '''
    prism_barycentric_coordinates(tri1, tri2, point) yields a list of weights for each vertex
      in the given tetrahedron in the same order as the vertices given. If all weights are 0, then
      the point is not inside the tetrahedron. The returned weights are (a,b,d) in a numpy array;
      the values a, b, and c are the barycentric coordinates corresponding to the three points of
      the triangles (where c = (1 - a - b) and the value d is the fractional distance (in the range
      [0,1]) of the point between tri1 (d=0) and tri2 (d=1).
    '''
    pt = np.asarray(pt)
    tri1 = np.asarray(tri1)
    tri2 = np.asarray(tri2)
    (tri1,tri2) = [
        (np.transpose(tri, (1,0) if len(tri.shape) == 2 else (2,0,1)) if tri.shape[0] != 3 else
         np.transpose(tri, (0,2,1))                                   if tri.shape[1] != 3 else
         tri)
        for tri in (tri1,tri2)]
    pt = pt.T if pt.shape[0] != 3 else pt
    ## if the triangles aren't on coherent sides of each other, something is wrong with this prism
    ## (for now we don't do this because I'm not convinced it's necessary with cortex)
    #faces   = np.asarray([tri1, tri2])
    #(u1,u2) = np.cross(faces[:,1] - faces[:,0], faces[:,2] - faces[:,0], axis=1)
    #(p1,p2) = np.reshape([tri1[0], tri2[0]], (2,1,3,-1))
    ## make sure all three points from both triangles are on the right side of the other
    #sd1to2  = np.sign(np.sum((tri2 - p1) * np.reshape(u1, (1,3,-1)), axis=1))
    #sd2to1  = np.sign(np.sum((tri1 - p2) * np.reshape(u2, (1,3,-1)), axis=1))
    #bad = np.where(#np.any(sd1to2 == sd2to1, axis=0) |
    #               (sd1to2[0] != sd1to2[1])         |
    #               (sd1to2[0] != sd1to2[2]))[0]
    # get the individual tetrahedron bc coordinates; we divide the prism up into a few tetrahedrons
    ((a1,b1,c1),(a2,b2,c2)) = (tri1, tri2)
    (ab2,bc2,ca2) = np.mean([(a2,b2), (b2,c2), (c2,a2)], axis=1)
    mu1 = np.mean(tri1, axis=0)
    tetpts = np.asarray([a1,b1,c1, a2,b2,c2, ab2,bc2,ca2, mu1])
    bcs = np.vstack([tetrahedral_barycentric_coordinates(tetpts[tet], pt)
                     for tet in prism_barycentric_coordinates.tetrahedrons])
    # add up the weights on the points...
    abc12 = np.zeros([6, bcs.shape[1]])
    for (k, ii, wii) in prism_barycentric_coordinates.weight_plan:
        for (i,wi) in zip(ii, wii):
            abc12[wi] += bcs[i] / k
    # okay, now separate into h and ab:
    abc = abc12[:3] + abc12[3:]
    tot = np.sum(abc, axis=0)
    abc[2] = 1 - np.sum(abc12[:3], axis=0)
    bad = np.where(~np.isclose(tot, 1))[0]
    abc[:,bad] = 0
    q = tot[(tot > 1) & ~np.isclose(tot, 1)]
    return abc
# meta-data used by the above function:
# order of points is a1, b1, c1,  a2, b2, c2,  ab2, bc2, ca2,  mu1, where 1 is the first triangle,
# 2 is the second trignale, uv1 is the average of points u1 and v1, and mu1 is the average of tri1.
prism_barycentric_coordinates.weight_counts = np.asarray([1,1,1, 1,1,1, 2,2,2, 3])
prism_barycentric_coordinates.weights = {
    1: np.reshape(np.arange(6), (6,1)),
    2: np.array([[-1,-1],[-1,-1],[-1,-1],[-1,-1],[-1,-1],[-1,-1], # a1-c3 (0-5)
                 [3,4], [4,5], [5,3]]), #ab2-ca2 (6-8)
    3: np.full((10, 3), -1)}
prism_barycentric_coordinates.weights[3][-1] = [0,1,2]
prism_barycentric_coordinates.tetrahedrons = np.array([[0, 1, 6, 9],  [1, 2, 7, 9],  [2, 0, 8, 9],
                                                       [0, 6, 8, 9],  [1, 6, 7, 9],  [2, 7, 8, 9],
                                                       [0, 6, 8, 3],  [1, 6, 7, 4],  [2, 7, 8, 5],
                                                       [6, 7, 8, 9]])
prism_barycentric_coordinates.tetflat = prism_barycentric_coordinates.tetrahedrons.flatten()
prism_barycentric_coordinates.tetflat_weight_count = prism_barycentric_coordinates.weight_counts[
    prism_barycentric_coordinates.tetflat]
prism_barycentric_coordinates.weight_plan = tuple(
    [(k, ii, v[prism_barycentric_coordinates.tetflat[ii]])
     for (k,v) in six.iteritems(prism_barycentric_coordinates.weights)
     for ii in np.where(prism_barycentric_coordinates.tetflat_weight_count == k)])

def point_in_prism(tri1, tri2, pt):
    '''
    point_in_prism(tri1, tri2, pt) yields True if the given point is inside the prism that stretches
      between triangle 1 and triangle 2. Will automatically thread over extended dimensions. If
      multiple triangles are given, then the vertices must be an earlier dimension than the
      coordinates; e.g., a 3 x 3 x n array will be assumed to organized such that element [0,1,k] is
      the y coordinate of the first vertex of the k'th triangle.
    '''
    bcs = prism_barycentric_coordinates(tri1, tri2, pt)
    return np.logical_not(np.isclose(np.sum(bcs, axis=0), 0))



####################################################################################################
# neuropythy/math/__init__.py
# The Neuropythy math core module.
# by Noah C. Benson

from .core import (pi, half_pi, quarter_pi, tau, inf, nan, radperdeg, degperrad,
                   pytorch, to_torchdtype, torchdtype_to_numpydtype,
                   isarray, istensor, issparse, isdense, eq, ne, le, lt, ge, gt,
                   clone, totensor, astensor, toarray, asarray, asdense, promote, reshape,
                   add, sub, mul, div, mod, safesqrt, sqrt, exp, log, log10, log2, abs,
                   arcsin, sin, arccos, cos, arctan, tan, lgamma, hypot, hypot2,
                   triarea, eudist2, eudist, trisides2, trisides, trialtitudes,
                   rangemod, radmod, degmod, branch, zinv, numel, rows, check_sparsity, unbroadcast,
                   sum, prod, mean, var, std, median, min, max, argmin, argmax, all,
                   beta_log_prob, beta_prob, normal_log_prob, normal_prob,
                   cauchy_log_prob, cauchy_prob, halfcauchy_log_prob, halfcauchy_prob, 
                   laplace_log_prob, laplace_prob, exp_log_prob, exp_prob,
                   gennorm_log_prob, gennorm_prob, gumbel_log_prob, gumbel_prob)

pytorch_ok = False
try:
    if pytorch() is not None:
        pytorch_ok = True
except ImportError: pass




####################################################################################################
# neuropythy/geometry/__init__.py
# This file defines common rotation functions that are useful with cortical mesh spheres, such as
# those produced with FreeSurfer.

'''
The neuropythy.geometry package contains a number of utility functions for calculating 2D and 3D
geometrical values as well as three classes: Mesh, Tesselation, and Topology. The Tesselation
class stores information about triangle meshes---essentially all the information except the 2D or 3D
coordinates of the vertices and that information which requires those coordinates. The Mesh class
simply reifies the Tesselation class with these coordinates and the relevant values that can be
derived from them. Finally, the Topology class tracks a Tesselation object and a set of Meshes that
share that tesselation in common.
'''

from .util import (
    normalize,
    vector_angle_cos,
    vector_angle,
    spherical_distance,
    rotation_matrix_3D,
    rotation_matrix_2D,
    alignment_matrix_3D,
    alignment_matrix_2D,
    point_on_line, point_on_segment, point_in_segment, points_close,
    lines_colinear, segments_colinear, segments_overlapping,
    lines_touch_2D, segments_touch_2D,
    line_intersection_2D,
    segment_intersection_2D,
    line_segment_intersection_2D,
    triangle_area,
    triangle_normal,
    cartesian_to_barycentric_2D,
    cartesian_to_barycentric_3D,
    barycentric_to_cartesian,
    triangle_address,
    triangle_unaddress,
    point_in_triangle,
    point_in_tetrahedron,
    point_in_prism,
    tetrahedral_barycentric_coordinates,
    prism_barycentric_coordinates)
from .mesh import (VertexSet, Tesselation, Mesh, Topology, MapProjection, Path, PathTrace,
                   mesh, is_mesh, is_flatmap,
                   tess, is_tess,
                   topo, is_topo,
                   is_vset, is_path, deduce_chirality,
                   map_projection, is_map_projection,
                   load_map_projection, load_projections_from_path,
                   projections_path, map_projections, 
                   path_trace, is_path_trace, close_path_traces, paths_to_labels,
                   to_tess, to_mesh, to_property, to_mask, isolines, smooth_lines,
                   to_map_projection, to_flatmap)




####################################################################################################
# neuropythy/commands/atlas.py
# The code for the function that handle the application of atlases to the cortical surface of a
# subject.
# By Noah C. Benson

from __future__ import print_function

import os, sys, six, re, pimms, textwrap, warnings
import numpy                        as     np
import scipy                        as     sp
import nibabel                      as     nib
import nibabel.freesurfer.io        as     fsio
import nibabel.freesurfer.mghformat as     fsmgh
import pyrsistent                   as     pyr

from   ..freesurfer                 import (subject as freesurfer_subject)
from   ..hcp                        import (subject as hcp_subject)
from   ..util                       import (curry, library_path, AutoDict)
from   ..util.conf                  import config
from   ..                           import io as nyio

@pimms.calc('worklog')
def calc_worklog(stdout=Ellipsis, stderr=Ellipsis, verbose=False):
    '''
    calc_worklog constructs the worklog from the stdout, stderr, stdin, and verbose arguments.
    '''
    try: cols = int(os.environ['COLUMNS'])
    except Exception: cols = 80
    return pimms.worklog(columns=cols, stdout=stdout, stderr=stderr, verbose=verbose)
@pimms.calc('subject')
def calc_subject(argv, worklog):
    '''
    calc_subject converts a subject_id into a subject object.

    Afferent parameters:
      @ argv
        The FreeSurfer subject name(s), HCP subject ID(s), or path(s) of the subject(s) to which the
        atlas should be applied.
    '''
    if len(argv) == 0: raise ValueError('No subject-id given')
    elif len(argv) > 1: worklog.warn('WARNING: Unused subject arguments: %s' % (argv[1:],))
    subject_id = argv[0]
    try:
        sub = freesurfer_subject(subject_id)
        if sub is not None:
            worklog('Using FreeSurfer subject: %s' % sub.path)
            return sub
    except Exception: pass
    try:
        sub = hcp_subject(subject_id)
        if sub is not None:
            worklog('Using HCP subject: %s' % sub.path)
            return sub
    except Exception: pass
    raise ValueError('Could not load subject %s' % subject_id)
@pimms.calc('atlas_map', 'atlas_subject')
def calc_atlases(worklog, atlas_subject_id='fsaverage'):
    '''
    calc_atlases finds all available atlases in the possible subject directories of the given atlas
    subject.

    In order to be a template, it must either be a collection of files (either mgh/mgz or FreeSurfer
    curv/morph-data files) named as '<hemi>.<template>_<quantity><ending>' such as the files
    'lh.wang2015_mplbl.mgz' and 'rh.wang2015_mplbl.mgz'. They may additionally have a version
    prior to the ending, as in 'lh.benson14_angle.v2_5.mgz'. Files without versions are considered 
    to be of a higher version than all versioned files. All files must be found in the atlas
    subject's surf/ directory; however, all subjects in all FreeSurfer subjects paths with the same
    subject id are searched if the atlas is not found in the atlas subejct's directory.

    Afferent parameters:
      @ atlas_subject_id 
        The FreeSurfer subject name subject path of the subject that is to be used as the atlas
        subject from which the atlas is interpolated. HCP subjects are not currently supported.

    Efferent values:
      @ atlas_map 
        A persistent map whose keys are atlas names, the values of which are themselves persistent
        maps whose keys are the versions of the given atlas (None potentially being included). The
        values of these maps are again maps of hemisphere names then finally of the of the quantity
        names (such as 'eccen' or 'maxprob') to the property vectors imported from the appropriate
        files.
    '''
    try:              sub = freesurfer_subject(atlas_subject_id)
    except Exception: sub = None
    if sub is None:
        try: sub = hcp_subject(atlas_subject_id)
        except Exception: sub = None
    if sub is None: raise ValueError('Could not load atlas subject %s' % atlas_subject_id)
    worklog('Using Atlas subject: %s' % sub.path)
    # Now find the requested atlases
    atlases = AutoDict()
    atlas_patt = r'^([lr]h)\.([^_]+)_([^.]+)(\.(v(\d+(_\d+)*)))?((\.mg[hz])|\.nii(\.gz)?)?$'
    atlas_hemi_ii = 1
    atlas_atls_ii = 2
    atlas_meas_ii = 3
    atlas_vrsn_ii = 6
    libdir = os.path.join(library_path(), 'data')
    for pth in [libdir] + config['freesurfer_subject_paths'] + [sub.path]:
        # see if appropriate files are in this directory
        pth = os.path.join(pth, sub.name, 'surf')
        if not os.path.isdir(pth): continue
        for fl in os.listdir(pth):
            m = re.match(atlas_patt, fl)
            if m is None: continue
            fl = os.path.join(pth, fl)
            (h, atls, meas, vrsn) = [
                m.group(ii) for ii in (atlas_hemi_ii, atlas_atls_ii, atlas_meas_ii, atlas_vrsn_ii)]
            if vrsn is not None: vrsn = tuple([int(s) for s in vrsn.split('_')])
            atlases[atls][vrsn][h][meas] = curry(nyio.load, fl)
    # convert the possible atlas maps into persistent/lazy maps
    atlas_map = pyr.pmap({a:pyr.pmap({v:pyr.pmap({h:pimms.lazy_map(hv)
                                                  for (h,hv) in six.iteritems(vv)})
                                      for (v,vv) in six.iteritems(av)})
                          for (a,av) in six.iteritems(atlases)})
    return {'atlas_map':atlas_map, 'atlas_subject':sub}
@pimms.calc('subject_cortices', 'atlas_cortices')
def calc_cortices(subject, atlas_subject, worklog, hemis=None):
    '''
    calc_cortices extracts the hemisphere objects (of the subject) to which the atlas is being
    applied. By default these are 'lh' and 'rh', but for HCP subjects other hemispheres may be
    desired.

    Afferent parameters:
      @ hemis 
        The hemispheres onto which to put the atlas; this may take a number of forms: 'lh' or 'rh' 
        (or --hemis=lh / --hemis=rh) applies the atlas to the given hemisphere only; otherwise a
        list of hemispheres may be specified as a python object (e.g.
        --hemis='("lh_LR32k","rh_LR32k")') or as a comma or whitespace separated string (such as
        'lh_LR32k,lh_LR59k' or 'rh rh_LR164k'). 'lr', 'both', and 'all' are equivalent to 'lh rh';
        this is the default behavior if hemis is not provided explicitly.
      @ subject 
        The neuropythy subject object onto which the atlas is being projected.
      @ atlas_subject 
        Theneuropythy subject object from which the atlas is being projected.
    '''
    if hemis is None or hemis is Ellipsis: hemis = 'lr'
    if pimms.is_str(hemis):
        if hemis.lower() in ['lr', 'both', 'all']: hemis = ('lh','rh')
        else: hemis = re.split(r'([,;:]|\s)+', hemis)[::2]
    if atlas_subject.name == 'fsaverage_sym':
        hemis = ['rhx' if h == 'rh' else h for h in hemis]
    sctcs = {}
    actcs = {}
    worklog('Preparing Hemispheres...')
    for h in hemis:
        if h not in subject.hemis:
            raise ValueError('Subject %s does not have requested hemi %s' % (subject.name, h))
        sctcs[h] = curry(lambda sub,h: sub.hemis[h], subject,       h)
        h = 'lh' if h == 'rhx' else h
        if h not in atlas_subject.hemis:
            raise ValueError('Atlas subject %s does not have requested hemi %s' % (
                atlas_subject.name, h))
        actcs[h] = curry(lambda sub,h: sub.hemis[h], atlas_subject, h)
    return {'subject_cortices': pimms.lazy_map(sctcs),
            'atlas_cortices':   pimms.lazy_map(actcs)}
@pimms.calc('atlas_properties', 'atlas_version_tags')
def calc_atlas_projections(subject_cortices, atlas_cortices, atlas_map, worklog, atlases=Ellipsis):
    '''
    calc_atlas_projections calculates the lazy map of atlas projections.

    Afferent parameters:
      @ atlases 
        The atlases that should be applied to the subject. This can be specified as a list/tuple of
        atlas names or as a string where the atlas names are separated by whitespace, commas, or
        semicolons. For example, to specify the 'benson14' atlas as well as the 'wang15' atlas, then
        ('benson14', 'wang15'), 'benson14 wang15' or 'benson14,wang15' would all be acceptable. To
        specify an atlas version, separate the atlas-name and the version with a colon (:), such as
        'benson14:2.5'. If no version is provided, then the highest version found is used. If
        atlases is set to None or Ellipsis (the default), this is equivalent to 'benson14,wang15'.

    Efferent values:
      @ atlas_properties 
        The atlas properties is a nested pimms lazy map whose key-path are like those of the
        atlas_map afferent parameter but which contains only those atlases requested via the atlases
        afferent parameter and whose deepest values are interpolated property vectors for the 
        target subject.
      @ atlas_version_tags 
        Each atlas can be specified as <atlas> or <atlas>:<version>; if the version is specified,
        then the version tag string (e.g., '.v1_5') is included in this dictionary; if only <atlas>
        was specified then this string is ''. If <atlas>: is specified, then the version string for
        whichever atlas was used is included.
    '''
    # Parse the atlases argument first:
    if atlases is Ellipsis:   atlases = ('benson14', 'wang15')
    if pimms.is_str(atlases): atlases = tuple(re.split(r'([,;]|\s)+', atlases)[::2])
    def _atlas_to_atlver(atl):
        atl0 = atl
        if not pimms.is_vector(atl):
            if ':' in atl:
                atl = atl.split(':')
                if len(atl) != 2: raise ValueError('Cannot parse atlas spec: %s' % atl0)
            else: atl = [atl, None]
        if len(atl) != 2: raise ValueError('Improperly specified atlas: %s' % atl0)
        if pimms.is_str(atl[1]):
            if len(atl[1]) == 0: atl = (atl[0], None)
            else:
                if atl[1][0] == 'v': atl[1] = atl[1][1:]
                try: atl = (atl[0], tuple([int(x) for x in re.split(r'[-_.]+', atl[1])]))
                except Exception:
                    raise ValueError('Could not parse atlas version string: %s' % atl[1])
        elif pimms.is_int(atl[1]):  atl = (atl[0], (atl[1],))
        elif pimms.is_real(atl[1]): atl = (atl[0], (int(atl[1]), int(10*(atl[1] - int(atl[1]))),))
        elif pimms.is_vector(atl[1], int): atl = (atl[0], tuple(atl[1]))
        elif atl[1] is not None:
            raise ValueError('atlas version must be a string (like "v1_5_1") or a list of ints')
        else: atl = tuple(atl)
        return atl + (atl0,)
    # Okay, let's find these versions of the atlases in the atlas_map...
    worklog('Preparing Atlases...')
    wl = worklog.indent()
    atl_props = AutoDict()
    avt = AutoDict()
    # keyfn is for sorting versions (newest version last)
    keyfn = lambda k:((np.inf,)     if k is None                 else
                      k + (np.inf,) if len(k) == 0 or k[-1] != 0 else 
                      k)
    for (atl,version,atl0) in [_atlas_to_atlver(atl) for atl in atlases]:
        if atl not in atlas_map: raise ValueError('Could not find an atlas named %s' % atl)
        atldat = atlas_map[atl]
        # if the version is None, we pick the highest of the available versions
        if version is None: v = sorted(atldat.keys(), key=keyfn)[-1]
        elif version in atldat: v = version
        else: raise ValueError('Could not find specific version %s of atlas %s' % (version,atl))
        # update the atlas-version-tag data
        wl('Atlas: %s, Version: %s' % (atl, v))
        avt[atl][v] = '' if v is None or ':' not in atl0 else ('.v' + '_'.join(map(str, v)))
        lmaps = atlas_map[atl][v]
        # convert these maps into interpolated properties...
        for (h,hmap) in six.iteritems(lmaps):
            hmap = pimms.lazy_map(
                {m: curry(
                    lambda hmap,h,m: atlas_cortices[h].interpolate(
                        subject_cortices[h],
                        hmap[m]),
                    hmap, h, m)
                 for m in six.iterkeys(hmap)})
            lmaps = lmaps.set(h, hmap)
        # add the lmaps (the persistent/lazy maps for this atlas version) in the atlprops
        atl_props[atl][v] = lmaps
    # That's all; we can return atl_props once we persist it
    return {'atlas_properties':  pimms.persist(atl_props),
            'atlas_version_tags': pimms.persist(avt)}
@pimms.calc('atlas_images', 'image_template_object')
def calc_images(subject, atlas_properties, image_template, worklog, volume_export=False):
    '''
    calc_images is a calculator that converts the atlas properties into a similar nested map of
    atlas images.

    Afferent parameters:
      @ image_template 
        The image template object or file to be used as a template for the output image. This
        can be None (in which case 'brain' is used) or a string that identifies an image of the
        subject. Or it can be a valid filename fo an image.

    Efferent values:
      @ atlas_images 
        A nested lazy-map structure that contains the 3D images for each of the relevant atlases
        and properties.
    '''
    from neuropythy import image_clear
    # parse the image template
    if not volume_export:
        return {'atlas_images': {}, 'image_template_object': None}
    if image_template is None:
        k = next((k for k in ['brain', 'ribbon', 'original']
                  if k in subject.images and subject.images.get(k, None) is not None),
                 None)
        if k is None: raise ValueError('Could not find a valid default image template')
        else: image_template = subject.images[k]
    elif pimms.is_str(image_template):
        if image_template in subject.images: 
            image_template = subject.images[image_template]
        else:
            try: image_template = ny.load(image_template, to='image')
            except Exception: image_template = None
    if image_template is None:
        raise ValueError('Could not load or find image template: %s' % (image_template,))
    image_template = image_clear(image_template)
    # having the image template, make the addresses:
    hemis = list(six.iterkeys(next(six.itervalues(next(six.itervalues(atlas_properties))))))
    addr = pimms.lmap({h: curry(lambda h: subject.hemis[h].image_address(image_template), h)
                       for h in hemis})
    worklog('Preparing Images...')
    def _make_images(vd, m):
        #worklog('Constructing %s images...' % (m,))
        dat = {h: vd[h][m] for h in six.iterkeys(vd)}
        lk = next(iter(sorted(k for k in six.iterkeys(dat) if k.lower().startswith('lh'))), None)
        rk = next(iter(sorted(k for k in six.iterkeys(dat) if k.lower().startswith('rh'))), None)
        idat = (dat[lk] if lk is not None else None, dat[rk] if rk is not None else None)
        if np.array_equal(idat, (None, None)): return None
        aa = (addr.get(lk, None), addr.get(rk, None))
        im = subject.cortex_to_image(idat, image_template, hemi=(lk, rk), address=aa)
        return im
    ims = {atl: {v: pimms.lmap({m: curry(_make_images, vd, m) for m in six.iterkeys(hd)})
                 for (v,vd) in six.iteritems(pps)
                 for hd in [next(six.itervalues(vd))]}
           for (atl, pps) in six.iteritems(atlas_properties)}
    return {'atlas_images': pimms.persist(ims),
            'image_template_object': image_template}
@pimms.calc('filemap', 'export_all_fn')
def calc_filemap(atlas_properties, atlas_images, subject, atlas_version_tags, worklog,
                 image_template_object, volume_export=False, surface_export=True, output_path=None,
                 volume_path=None, overwrite=False, output_format='mgz', volume_format='mgz',
                 create_directory=False):
    '''
    calc_filemap is a calculator that converts the atlas properties nested-map into a single-depth
    map whose keys are filenames and whose values are the interpolated property data.

    Afferent parameters
      @ output_path 
        The directory into which the atlas files should be written. If not provided or None then
        uses the subject's surf directory. If this directory doesn't exist, then it uses the
        subject's directory itself.
      @ volume_path 
        The directory into which the volume files should be written. If not provided or None then
        uses the subject's mri directory. If this directory doesn't exist, then it uses the
        subject's directory itself.
      @ overwrite 
        Whether to overwrite existing atlas files. If True, then atlas files that already exist will
        be overwritten. If False, then no files are overwritten.
      @ create_directory 
        Whether to create the output path if it doesn't exist. This is False by default.
      @ output_format 
        The desired output format of the files to be written. May be one of the following: 'mgz',
        'mgh', or either 'curv' or 'morph'.
      @ volume_export 
        Whether to include the volume files in the export_all_fn; default is False.
      @ surface_export 
        Whether to include the surface files in the export_all_fn; default if True.

    Efferent values:
      @ filemap 
        A pimms lazy map whose keys are filenames and whose values are interpolated atlas
        properties.
      @ export_all_fn 
        A function of no arguments that, when called, exports all of the files in the filemap to the
        output_path.
    '''
    if output_path is None:
        output_path = os.path.join(subject.path, 'surf')
        if not os.path.isdir(output_path): output_path = subject.path
    if volume_export and volume_path is None:
        volume_path = os.path.join(subject.path, 'mri')
        if not os.path.isdir(volume_path): volume_path = subject.path
    output_format = 'mgz' if output_format is None else output_format.lower()
    if output_format.startswith('.'): output_format = output_format[1:]
    (fmt,ending) = (('mgh','.mgz') if output_format == 'mgz' else
                    ('mgh','.mgh') if output_format == 'mgh' else
                    ('freesurfer_morph',''))
    if volume_export:
        imtype = type(image_template_object)
        if imtype is nib.Nifti1Image or imtype is nib.Nifti2Image:
            imending = '.nii.gz'
        elif imtype is nib.MGHImage:
            imending = '.mgz'
        else:
            imending = '.mgz'
            worklog('Note: Using image type MGH/MGZ in absence of explicit template')
    # make the filemap...
    worklog('Preparing Filemap...')
    fm = AutoDict()
    sfiles = []
    vfiles = []
    for (atl,atldat) in six.iteritems(atlas_properties):
        for (ver,verdat) in six.iteritems(atldat):
            vstr = atlas_version_tags[atl][ver]
            for (h,hdat) in six.iteritems(verdat):
                for m in six.iterkeys(hdat):
                    flnm = '%s.%s_%s%s%s' % (h, atl, m, vstr, ending)
                    flnm = os.path.join(output_path, flnm)
                    fm[flnm] = curry(lambda hdat,m: hdat[m], hdat, m)
                    sfiles.append(flnm)
            # now volumes...
            if volume_export:
                ims = atlas_images[atl][ver]
                for m in six.iterkeys(ims):
                    flnm = '%s_%s%s%s' % (atl, m, vstr, imending)
                    flnm = os.path.join(volume_path, flnm)
                    fm[flnm] = curry(lambda ims,m: ims[m], ims, m)
                    vfiles.append(flnm)
    # okay, make that a lazy map:
    filemap = pimms.lazy_map(fm)
    # the function for exporting all properties:
    def export_all():
        '''
        This function will export all files from its associated filemap and return a list of the
        filenames.
        '''
        if not os.path.isdir(output_path):
            if not create_directory:
                raise ValueError('No such path and create_direcotry is False: %s' % output_path)
            os.makedirs(os.path.abspath(output_path), 0o755)
        filenames = []
        worklog('Extracting Files...')
        wl = worklog.indent()
        if surface_export:
            for flnm in sfiles:
                wl(flnm)
                filenames.append(nyio.save(flnm, filemap[flnm], fmt))
        if volume_export:
            for flnm in vfiles:
                wl(flnm)
                filenames.append(nyio.save(flnm, filemap[flnm]))
        return filenames
    return {'filemap': filemap, 'export_all_fn': export_all}

atlas_plan_data = pyr.pmap(
    {'init_worklog':calc_worklog,
     'init_subject':calc_subject,
     'init_atlases':calc_atlases,
     'init_cortices':calc_cortices,
     'atlas_properties':calc_atlas_projections,
     'atlas_images':calc_images,
     'filemap':calc_filemap})
atlas_plan = pimms.plan(atlas_plan_data)

atlas_cmdline_abbrevs = {'output_format':    'f',
                         'volume_format':    'F',
                         'atlases':          'a',
                         'overwrite':        'o',
                         'create_directory': 'c',
                         'subject_id':       's',
                         'hemis':            'H',
                         'output_path':      'o',
                         'volume_path':      'O',
                         'atlas_subject_id': 'r',
                         'image_template':   'i',
                         'surface_export':   's',
                         'volume_export':    'S',
                         'verbose':          'v'}

def _format_afferent_doc(docstr, abbrevs=None, cols=80):
    try:
        (ln1, docs) = docstr.split('\n\n')
    except Exception: return ''
    anm0 = ln1.split(' (')[0]
    anm = anm0.replace('_', '-')
    header = '  --' + anm
    if abbrevs and anm0 in abbrevs: header = header + ' | -' + abbrevs[anm0]
    docs = [
        '\n     '.join(
            textwrap.wrap(
                ' '.join([s.strip() for s in ss.split('\n')[1:]]),
                cols-6))
        for ss in docs.split('\n)')]
    return header + '\n' + ''.join(['   * ' + d for d in docs])

info = \
    '''SYNTAX:  python -m neuropythy atlas <subject-id>

Neuropythy's atlas command interpolates atlases from the cortical surface of one
subject (usually an average subject such as fsaverage) onto that of another
subject. Most commonly this is used to see the anatomically-based prediction of
an ROI label or a parameter map (such as a retinotopic map) on an individual
subject.

The following optional arguments may be provided:
  --help | -h
    * Prints this help message.
''' + '\n'.join(
    [s for s in [_format_afferent_doc(atlas_plan.afferent_docs[af], atlas_cmdline_abbrevs)
                 for af in atlas_plan.afferents
                 if af in atlas_plan.afferent_docs]
     if len(s) > 0])

def main(*argv):
    '''
    neuropythy.commands.atlas.main() runs the main function of the atlas command in the neuropythy
      library. See the `python -m neuropythy atlas --help` or atlas.info for more information.
    '''
    argv = [aa for arg in argv for aa in (arg if pimms.is_vector(arg) else [arg])]
    imap = pimms.argv_parse(
        atlas_plan, argv,
        arg_abbrevs=atlas_cmdline_abbrevs)
    argv = imap['argv']
    if len(argv) == 0 or '--help' in argv or '-h' in argv:
        print(info)
        return 1
    try: imap['export_all_fn']()
    except Exception as e:
        raise
        sys.stderr.write('\nERROR:\n' + str(e) + '\n')
        sys.stderr.flush()
        sys.exit(2)
    return 0



####################################################################################################
# neuropythy/mri/images.py
# This file implements the core code used to manipulate and import/export images in neuropythy.
# by Noah C. Benson

import types, inspect, pimms, os, six, warnings, sys
import numpy                            as np
import scipy.sparse                     as sps
import pyrsistent                       as pyr
import nibabel                          as nib
import nibabel.freesurfer.mghformat     as fsmgh
from   functools                    import reduce

from ..util import (to_affine, is_image, is_image_header, is_tuple, curry, to_hemi_str, zinv)

# handy function for getting an image header
def to_image_header(img):
    '''
    to_image_header(img) yields img.header if img is a nibabel image object.
    to_image_header(hdr) yields hdr if hdr is a nibabel header object.
    to_image_header(obj) raises an error for other input types.
    '''
    if not img.__module__.startswith('nibabel.'):
        raise ValueError('to_image_header: only nibabel obejcts can be coerced to headers')
    if type(img).__name__.endswith('Header'): return img
    # if not a header given, must be an image given:
    try: return img.header
    except Exception:
        raise ValueError('to_image_header: can only convert nibabel image or header objects')

# ImageType Classes ################################################################################
# These classes define how images are interpreted and loaded; they are intended to be static.
class ImageType(object):
    '''
    The ImageType class defines how nibabel image types are handled by the neuropythy.mri.images
    subsystem.
    '''
    def __init__(self): raise NotImplementedError('ImageType is a static class')
    @classmethod
    def image_name(self): raise NotImplementedError('ImageType.image_name must be overloaded')
    @classmethod
    def image_type(self): raise NotImplementedError('ImageType.image_type must be overloaded')
    @classmethod
    def header_type(self): raise NotImplementedError('ImageType.header_type must be overloaded')
    @classmethod
    def aliases(self): return ()
    @classmethod
    def default_type(self): return np.dtype('float32')
    @classmethod
    def default_affine(self): return np.eye(4)
    @classmethod
    def parse_type(self, hdat, dataobj=None):
        '''
        Parses the dtype out of the header data or the array, depending on which is given; if both,
          then the header-data overrides the array; if neither, then np.float32.
        '''
        try: dataobj = dataobj.dataobj
        except Exception: pass
        dtype = np.asarray(dataobj).dtype if dataobj is not None else self.default_type()
        if   hdat and hdat.get('type', None) not in [None,Ellipsis]: dtype = np.dtype(hdat['type'])
        elif hdat and hdat.get('dtype',None) not in [None,Ellipsis]: dtype = np.dtype(hdat['dtype'])
        return dtype
    @classmethod
    def parse_affine(self, hdat, dataobj=None):
        '''
        Parses the affine out of the given header data and yields it.
        '''
        if 'affine' in hdat: return to_affine(hdat['affine'])
        else:                return to_affine(self.default_affine())
    @classmethod
    def parse_dataobj(self, dataobj, hdat={}):
        # first, see if we have a specified shape/size
        ish = next((hdat[k] for k in ('image_size', 'image_shape', 'shape') if k in hdat), None)
        if ish is Ellipsis: ish = None
        # make a numpy array of the appropriate dtype
        dtype = self.parse_type(hdat, dataobj=dataobj)
        try:    dataobj = dataobj.dataobj
        except Exception: pass
        if   dataobj is not None: arr = np.asarray(dataobj).astype(dtype)
        elif ish:                 arr = np.zeros(ish,       dtype=dtype)
        else:                     arr = np.zeros([1,1,1,0], dtype=dtype)
        # reshape to the requested shape if need-be
        if ish and ish != arr.shape: arr = np.reshape(arr, ish)
        # then reshape to a valid (4D) shape
        sh = arr.shape
        if   len(sh) == 2: arr = np.reshape(arr, (sh[0], 1, 1, sh[1]))
        elif len(sh) == 1: arr = np.reshape(arr, (sh[0], 1, 1))
        elif len(sh) == 3: arr = np.reshape(arr, sh)
        elif len(sh) != 4: raise ValueError('Cannot convert n-dimensional array to image if n > 4')
        # and return
        return arr
    @classmethod
    def parse_kwargs(self, arr, hdat={}):
        ext = hdat.get('extra', {})
        for (k,v) in six.iteritems(hdat):
            if k in ['header', 'extra', 'file_map']: continue
            ext[k] = v
        kw = {'extra': ext} if len(ext) > 0 else {}
        if 'header' in hdat: kw['header'] = hdat['header']
        if 'file_map' in hdat: kw['file_map'] = hdat['file_map']
        return kw
    @classmethod
    def to_image(self, arr, hdat={}):
        # reshape the data object or create it empty if None was given:
        arr = self.parse_dataobj(arr, hdat)
        # get the affine
        aff = self.parse_affine(hdat, dataobj=arr)
        # get the keyword arguments
        kw = self.parse_kwargs(arr, hdat)
        # create an image of the appropriate type
        cls = self.image_type()
        img = cls(arr, aff, **kw)
        # post-process the image
        return self.postprocess_image(img, hdat)
    @classmethod
    def postprocess_image(self, img, hdat={}): return img
    @classmethod
    def create(self, arr, meta_data={}, **kwargs):
        '''
        itype.create(dataobj) yields an image of the given image type itype that represents the
          given data object dataobj.
        itype.create(dataobj, meta_data) uses the given meta/header data to create the image.

        Any number of keyword arguments may also be appended to the call; these are merged into the
        meta_data argument.
        '''
        return self.to_image(arr, hdat=pimms.merge(meta_data, kwargs))
    @classmethod
    def meta_data(self, img):
        hdr = to_image_header(img)
        d = {}
        # basic stuff (most headers should have these)
        try: d['affine'] = img.affine
        except Exception:
            try:   d['affine'] = hdr.get_best_affine()
            except Exception: pass
        try: d['voxel_size'] = hdr.get_zooms()
        except Exception: pass
        try: d['voxel_type'] = hdr.get_data_dtype()
        except Exception: pass
        try: d['image_shape'] = hdr.get_data_shape()
        except Exception: pass
        # less basic stuff (some have these)
        try: d['bytes_per_voxel'] = hdr.get_data_bytespervox()
        except Exception: pass
        try: d['image_bytes'] = hdr.get_data_size()
        except Exception: pass
        try: d['image_offset'] = hdr.get_data_offset()
        except Exception: pass
        try:
            (m,b) = hdr.get_slope_inter()
            d['data_slope']  = m
            d['data_offset'] = b
        except Exception: pass
        # that's it
        return d
class MGHImageType(ImageType):
    @classmethod
    def image_name(self): return 'mgh'
    @classmethod
    def image_type(self): return fsmgh.MGHImage
    @classmethod
    def header_type(self): return fsmgh.MGHHeader
    @classmethod
    def default_affine(self): return to_affine([[-1, 0, 0, 0], [0, 0, 1, 0], [0, -1, 0, 0]], 3)
    @classmethod
    def default_type(self): return np.dtype('float32')
    @classmethod
    def aliases(self): return ('mgh', 'mgz', 'mgh.gz', 'freesurfer_image', 'freesurfer')
    @classmethod
    def parse_type(self, hdat, dataobj=None):
        dtype = super(MGHImageType, self).parse_type(hdat, dataobj=dataobj)
        if   np.issubdtype(dtype, np.floating): dtype = np.float32
        elif np.issubdtype(dtype, np.int8):     dtype = np.int8
        elif np.issubdtype(dtype, np.int16):    dtype = np.int16
        elif np.issubdtype(dtype, np.integer):  dtype = np.int32
        else: raise ValueError('Could not deduce appropriate MGH type for dtype %s' % dtype)
        return dtype
    @classmethod
    def meta_data(self, img):
        d = super(MGHImageType, self).meta_data(img)
        hdr = to_image_header(img)
        try:    d['vox2ras'] = hdr.get_vox2ras()
        except Exception: pass
        try:    d['ras2vox'] = hdr.get_ras2vox()
        except Exception: pass
        try:    d['vox2ras_tkr'] = hdr.get_vox2ras_tkr()
        except Exception: pass
        try:    d['footer_offset'] = hdr.get_footer_offset()
        except Exception: pass
        try:
            zooms = hdr.get_zooms()
            (zooms, tr) = (zooms, None) if len(zooms) == 3 else (zooms[:3], zooms[3])
            d['voxel_size']     = pimms.quant(zooms, 'mm')
            d['slice_duration'] = pimms.quant(tr, 'ms') if tr is not None else None
        except Exception: pass
        return d
class Nifti1ImageType(ImageType):
    @classmethod
    def image_name(self): return 'nifti1'
    @classmethod
    def image_type(self): return nib.Nifti1Image
    @classmethod
    def header_type(self): return nib.Nifti1Header
    @classmethod
    def aliases(self): return ('nii', 'nii.gz', 'nifti')
    @classmethod
    def meta_data(self, img):
        from nibabel.nifti1 import (slice_order_codes, unit_codes)
        d = super(Nifti1ImageType, self).meta_data(img)
        hdr = to_image_header(img)
        try: d['dimension_information'] = hdr.get_dim_info()
        except Exception: pass
        try: d['intent'] = hdr.get_intent()
        except Exception: pass
        try: d['slice_count'] = hdr.get_n_slices()
        except Exception: pass
        try: (sunit, tunit) = hdr.get_xyzt_units()
        except Exception: (sunit, tunit) = ('unknown', 'unknown')
        vsz = d['voxel_size']
        if vsz is not None and len(vsz) == 4:
            d['voxel_size'] = vsz[:3]
            if tunit == 'unknown': d['voxel_duration'] = vsz[3]
            else:
                try: d['voxel_duration'] = pimms.quant(vsz[3], tunit)
                except Exception: d['voxel_duration'] = vsz[3]
        else:
            d['voxel_duration'] = None
        if sunit != 'unknown':
            try: d['voxel_size'] = pimms.quant(d['voxel_size'], sunit)
            except Exception: pass
        try:
            sd = hdr.get_slice_duration()
            if tunit != 'unknown': sd = pimms.quant(sd, tunit)
            d['slice_duration'] = sd
        except Exception: pass
        try:
            (q,qc) = hdr.get_qform(True)
            d['qform_code'] = qc
            d['qform'] = q
        except Exception: pass
        try:
            (s,sc) = hdr.get_sform(True)
            d['sform_code'] = sc
            d['sform'] = s
        except Exception: pass
        try:
            sc = hdr['slice_code']
            if sc != 0: d['slice_order'] = slice_order_codes.label[sc]
        except Exception: pass
        try:
            ts = hdr.get_slice_times()
            ts = np.asarray([np.nan if t is None else t for t in ts])
            if tunit != 'unknown': ts = pimms.quant(ts, tunit)
            d['slice_times'] = ts
        except Exception: pass
        try:    d['header_size'] = hdr['sizeof_hdr']
        except Exception: pass
        try:    d['calibration'] = (hdr['cal_min'], hdr['cal_max'])
        except Exception: pass
        try:
            t0 = hdr['toffset']
            if tunits != 'unknown': t0 = pimms.quant(t0, tunits)
            d['time_offset'] = t0
        except Exception: pass
        try:    d['description'] = hdr['descrip']
        except Exception: pass
        try:    d['auxiliary_filename'] = hdr['aux_file']
        except Exception: pass
        return d
    @classmethod
    def unit_to_name(self, u):
        for name in ('meter', 'mm', 'sec', 'msec', 'usec'):
            if u == pimms.unit(name):
                return name
        # no current support for ppm designation
        if   u == pimms.unit('um'):     return 'micron'
        elif u == pimms.unit('1/sec'):  return 'hz'
        elif u == pimms.unit('Hz'):     return 'hz'
        elif u == pimms.unit('radian'): return 'rads'
        else: return 'unknown'
    @classmethod
    def postprocess_image(self, img, d):
        from nibabel.nifti1 import slice_order_codes
        hdr = img.header
        # dimension information:
        for k in ['dimension_information', 'dim_info', 'diminfo']:
            try:
                hdr.set_dim_info(*d[k])
                break
            except Exception: pass
        try:    hdr.set_intent(d['intent'])
        except Exception: pass
        # xyzt_units:
        try: sunit = self.unit_to_name(pimms.unit(d['voxel_size']))
        except Exception:
            try: sunit = self.unit_to_name(pimms.unit(d['voxel_unit']))
            except Exception: sunit = 'unknown'
        try: tunit = self.unit_to_name(pimms.unit(d['slice_duration']))
        except Exception:
            try: tunit = self.unit_to_name(pimms.unit(d['time_unit']))
            except Exception: tunit = 'unknown'
        try: hdr.set_xyzt_units(sunit, tunit)
        except Exception: pass
        # qform and sform
        try:
            try: q = to_affine(d['qform'])
            except Exception: q = to_affine(d['affine'])
            qc = d.get('qform_code', None)
            hdr.set_qform(q, qc)
        except Exception: pass
        try:
            try: s = to_affine(d['sform'])
            except Exception: s = to_affine(d['affine'])
            sc = d.get('sform_code', None)
            hdr.set_sform(s, sc)
        except Exception: pass
        # slice code
        try:    hdr['slice_code'] = slice_order_codes[d['slice_order']]
        except Exception: pass
        # slice duration
        try:
            dur = d['slice_duration']
            if pimms.is_quantity(dur):
                if tunit == 'unknown': dur = pimms.mag(dur)
                else: dur = pimms.mag(dur, tunit)
            hdr.set_slice_duration(dur)
        except Exception: pass
        # slice timing
        try:
            ts = d['slice_times']
            if pimms.is_quantity(ts):
                if tunit == 'unknown': ts = pimms.mag(ts)
                else: ts = pimms.mag(ts, tunit)
            hdr.set_slice_duration([None if np.isnan(t) else t for t in ts])
        except Exception: pass
        # slope / intercept
        try:    hdr.set_slope_inter(d.get('data_slope', None), d.get('data_offset', None))
        except Exception: pass
        # calibration
        try:
            (cmn, cmx) = d['calibration']
            hdr['cal_min'] = cmn
            hdr['cal_max'] = cmx
        except Exception: pass
        # time offset
        try:
            t0 = d['time_offset']
            if pimms.is_quantity(t0):
                if tunits != 'unknown': t0 = pimms.mag(t0, tunits)
                else: t0 = pimms.mag(t0)
            hdr['toffset'] = t0
        except Exception: pass
        # description
        try:    hdr['descrip'] = d['description']
        except Exception: pass
        # auxiliary filename
        try:    hdr['aux_file'] = d['auxiliary_filename']
        except Exception: pass
        return img
class Nifti2ImageType(Nifti1ImageType):
    @classmethod
    def image_name(self): return 'nifti2'
    @classmethod
    def image_type(self): return nib.Nifti2Image
    @classmethod
    def header_type(self): return nib.Nifti2Header
    @classmethod
    def aliases(self): return ('nii2', 'nii2.gz')
class Spm99AnalyzeImageType(ImageType):
    @classmethod
    def image_name(self): return 'spm99analyze'
    @classmethod
    def image_type(self): return nib.Spm99AnalyzeImage
    @classmethod
    def header_type(self): return nib.Spm99AnalyzeHeader
class Spm2AnalyzeImageType(Spm99AnalyzeImageType):
    @classmethod
    def image_name(self): return 'analyze'
    @classmethod
    def image_type(self): return nib.Spm2AnalyzeImage
    @classmethod
    def header_type(self): return nib.Spm2AnalyzeHeader
    @classmethod
    def aliases(self): return ('spm2analyze')
class Minc1ImageType(ImageType):
    @classmethod
    def image_name(self): return 'minc1'
    @classmethod
    def image_type(self): return nib.minc1.Minc1Image
    @classmethod
    def header_type(self): return nib.minc1.Minc1Header
    @classmethod
    def aliases(self): return ()
class Minc2ImageType(ImageType):
    @classmethod
    def image_name(self): return 'minc'
    @classmethod
    def image_type(self): return nib.minc2.Minc2Image
    @classmethod
    def header_type(self): return nib.minc2.Minc2Header
    @classmethod
    def aliases(self): return ('minc2')
class PARRECImageType(ImageType):
    @classmethod
    def image_name(self): return 'parrec'
    @classmethod
    def image_type(self): return nib.parrec.PARRECImage
    @classmethod
    def header_type(self): return nib.parrec.PARRECHeader
    @classmethod
    def aliases(self): return ()
    @classmethod
    def meta_data(self, img):
        d = super(PARRECImageType, self).meta_data(img)
        hdr = to_image_header(img)
        try:
            (bvals,bvec) = hdr.get_bvals_bvecs()
            d['b_values'] = bvals
            d['b_vectors'] = bvecs
        except Exception: pass
        try:
            (m,b) = hdr.get_data_scaling()
            d['data_slope'] = m
            d['data_offset'] = b
        except Exception: pass
        # get_def(name)?
        try:    d['echo_train_length'] = hdr.get_echo_train_length()
        except Exception: pass
        try:    d['record_shape'] = hdr.get_record_shape()
        except Exception: pass
        try:    d['slice_orientation'] = hdr.get_slice_orientation()
        except Exception: pass
        try:    d['sorted_slice_indices'] = hdr.get_sorted_slice_indices()
        except Exception: pass
        try:    d['volume_labels'] = hdr.get_volume_labels()
        except Exception: pass
        try:    d['water_fat_shift'] = hdr.get_water_fat_shift()
        except Exception: pass
        return d
class EcatImageType(ImageType):
    @classmethod
    def image_name(self): return 'ecat'
    @classmethod
    def image_type(self): return nib.ecat.EcatImage
    @classmethod
    def header_type(self): return nib.ecat.EcatHeader
    @classmethod
    def aliases(self): return ()
    @classmethod
    def meta_data(self, img):
        d = super(EcatImageType, self).meta_data(img)
        hdr = to_image_header(img)
        try: d['filetype'] = hdr.get_filetype()
        except Exception: pass
        try: d['patient_orientation'] = hdr.get_patient_orient()
        except Exception: pass
        return d
class Cifti2ImageType(ImageType):
    @classmethod
    def image_name(self): return 'cifti2'
    @classmethod
    def image_type(self): return nib.Cifti2Image
    @classmethod
    def header_type(self): return nib.Cifti2Header
    @classmethod
    def aliases(self): return ('cii', 'cii.gz', 'cifti')
    @classmethod
    def meta_data(self, img):
        from neuropythy.hcp import cifti_axis_spec
        d = {}
        axdat = cifti_axis_spec(img)
        d['axes_data'] = axdat
        d['image_shape'] = tuple([ax.get('size', None) for ax in axdat])
        #d['voxel_type'] = img.get_data_dtype()
        return pimms.persist(d)
    #@classmethod
    #def postprocess_image(self, img, d):
    #    return img


image_types = (Nifti1ImageType, Nifti2ImageType, MGHImageType,
               Spm99AnalyzeImageType, Spm2AnalyzeImageType,
               Minc1ImageType, Minc2ImageType,
               PARRECImageType, EcatImageType, Cifti2ImageType)
image_types_by_name        = pyr.pmap({it.image_name():it  for it in image_types})
image_types_by_image_type  = pyr.pmap({it.image_type():it  for it in image_types})
image_types_by_header_type = pyr.pmap({it.header_type():it for it in image_types})

# image-specs are allowed to store things using a few different keywords, so here are accessor
# functions based on some data structs:
imspec_aliases = {'image_shape': ['image_size', 'shape', 'image_dimensions', 'image_dims'],
                  'affine':      ['affine_transform', 'affine_matrix'],
                  'voxel_type':  ['dtype', 'type'],
                  'voxel_size':  ['pixel_size', 'voxel_dims', 'voxel_dimensions']}
def imspec_lookup(imspec, k, default=None):
    '''
    imspec_lookup(imspec, k) yields the value associated with the key k in the mapping imspec; if k
      is not in imspec, then imspec alises are checked and the appropriate value is returned;
      otherwise None is returned.
    imspec_lookup(imspec, k, default) yields default if neither k not an alias cannot be found.
    '''
    k = k.lower()
    if k in imspec: return imspec[k]
    aliases = imspec_aliases.get(k, None)
    if aliases is None:
        for q in six.itervalues(imspec_aliases):
            if k in q:
                aliases = q
                break
    if aliases is None: return default
    for kk in aliases:
        if kk in imspec: return imspec[kk]
    return default
def to_image_type(image_type):
    '''
    to_image_type(image_type) yields an image-type class equivalent to the given image_type
      argument, which may be a type name or alias or an image or header object or class.    
    '''
    if image_type is None: return None
    if isinstance(image_type, type) and issubclass(image_type, ImageType): return image_type
    if pimms.is_str(image_type):
        image_type = image_type.lower()
        if image_type in image_types_by_name: return image_types_by_name[image_type]
        for it in image_types:
            if image_type in it.aliases(): return it
        raise ValueError('"%s" is not a valid image-type name or alias' % image_type)
    for x in (image_type, type(image_type)):
        try:    return image_types_by_image_type[x]
        except Exception: pass
        try:    return image_types_by_header_type[x]
        except Exception: pass
    raise ValueError('Unsupported image type: %s' % image_type)
def is_image_array(arr):
    '''
    is_image_array(arr) yields True if arr is a valid image array and False otherwise; in order to
      be a valid array, it must be a 1D, 2D, 3D or 4D array.
    '''
    return pimms.is_array(arr, None, (1,2,3,4))
def is_image_spec(imspec):
    '''
    is_image_spec(imspec) yields True if imspec is a map with the keys 'affine' and 'image_shape',
      otherwise yields False.
    '''
    return (pimms.is_map(imspec) and
            imspec_lookup(imspec, 'affine') is not None and
            imspec_lookup(imspec, 'image_shape') is not None)
def image_header_to_spec(hdr):
    '''
    image_header_to_spec(hdr) yields an image spec for the given image header object hdr.
    '''
    intype = to_image_type(hdr)
    return intype.meta_data(hdr)
def image_shape(arg):
    '''
    image_shape(im) yields the image shape for the given image im. The argument im may be an image,
      an array, an image header, or an image spec.
    '''
    if   is_image(arg):                                sh = arg.shape
    elif pimms.is_vector(arg, 'int') and len(arg) < 5: sh = tuple(arg)
    elif is_image_spec(arg):                           sh = imspec_lookup(arg, 'image_shape')
    elif is_image_header(arg):                         sh = image_header_to_spec(arg)['image_shape']
    elif is_image_array(arg):                          sh = np.shape(arg)
    else: raise VelueError('Bad argument of type %s given to image_shape()' % type(arg))
    sh = tuple(sh)
    if   len(sh) == 2: sh = (sh[0], 1, 1, sh[1])
    elif len(sh) == 1: sh = (sh[0], 1, 1)
    return sh
def image_array_to_spec(arr):
    '''
    image_array_to_spec(arr) yields an image-spec that is appropriate for the given array. The 
      default image spec for an array is a FreeSurfer-like affine transformation with a translation
      that puts the origin at the center of the array. The upper-right 3x3 matrix for this
      transformation is [[-1,0,0], [0,0,1], [0,-1,0]].
    image_array_to_spec((i,j,k)) uses (i,j,k) as the shape of the image array.
    image_array_to_spec(image) uses the array from the given image but not the affine matrix.
    image_array_to_spec(spec) uses the image shape from the given image spec but not the affine
      matrix.
    '''
    sh   = image_shape(arr)[:3]
    (i0,j0,k0) = np.asarray(sh) * 0.5
    ijk0 = (i0, -k0, j0)
    aff  = to_affine(([[-1,0,0],[0,0,1],[0,-1,0]], ijk0), 3)
    return {'image_shape':sh, 'affine':aff}
def image_to_spec(img):
    '''
    image_to_spec(img) yields an image-spec for the given image img. If img is not an image object,
      this will fail.
    '''
    return image_header_to_spec(img.header)
def image_spec_to_image(imspec, image_type=None, fill=0):
    '''
    image_spec_to_image(imspec) yields an empty with the given image-spec.
    image_spec_to_image(imspec, image_type) creates an image with the given type
    image_spec_to_image(imspec, image_type, fill) fills the resulting image with the given
      fill-value.
    '''
    imsh = imspec_lookup(imspec, 'image_shape')
    if fill == 0: imarr = np.zeros(imsh, dtype=imspec_lookup(imspec, 'voxel_type'))
    else:         imarr = np.full(imsh, fill, dtype=imspec_lookup(imspec, 'voxel_type'))
    # okay, we have the image array...
    image_type = to_image_type('nifti1' if image_type is None else image_type)
    return image_type.create(imarr, meta_data=imspec)
def to_image_spec(img, **kw):
    '''
    to_image_spec(img) yields a dictionary of meta-data for the given nibabel image object img.
    to_image_spec(hdr) yields the equivalent meta-data for the given nibabel image header.

    Note that obj may also be a mapping object, in which case it is returned verbatim.
    '''
    if pimms.is_vector(img,'int') and is_tuple(img) and len(img) < 5:
        r = image_array_to_spec(np.zeros(img))
    elif pimms.is_map(img):    r = img
    elif is_image_header(img): r = image_header_to_spec(img)
    elif is_image(img):        r = image_to_spec(img)
    elif is_image_array(img):  r = image_array_to_spec(img)
    else: raise ValueError('cannot convert object of type %s to image-spec' % type(img))
    if len(kw) > 0: r = {k:v for m in (r,kw) for (k,v) in six.iteritems(m)}
    # normalize the entries
    for (k,aliases) in six.iteritems(imspec_aliases):
        if k in r: continue
        for al in aliases:
            if al in r:
                val = r[al]
                r = pimms.assoc(pimms.dissoc(r, al), k, val)
                break
    return r
def to_image(img, image_type=None, spec=None, **kwargs):
    '''
    to_image(array) yields a Nifti1Image of the given array with default meta-data spec.
    to_image(array, image_type) yields an image object of the given type; image_type may either be
      an image class or a class name (see supported types below).
    to_image((array, spec)) uses the given mapping of meta-data (spec) to construct the image-spec
      note that spec may simply be an affine transformation matrix or may be an image.
    to_image((array, affine, spec)) uses the given affine specifically (the given affine
      overrides any affine included in the spec meta-data).
    to_image(imspec) constructs an image with the properties specified in the given imspec; the
      special optional argument fill (default: 0.0) can be set to something else to specify what the
      default cell value should be.

    Note that the array may optionally be an image itself, in which case its spec is used as a
    starting point for the new spec. Any spec-data passed as a tuple overwrites this spec-data,
    and any spec-data passed as an optional argument overwrites this spec-data in turn.

    The first optional argument, specifying image_type is as an image type if possible, but if a
    spec-data mapping or equivalent (e.g., an image header or affine) is passed as the first
    argument it is used as such; otherwise, the optional third argument is named spec, and any
    additional keyword arguments passed to to_image are merged into this spec object left-to-right
    (i.e., keyword arguments overwrite the spec keys).

    If no affine is given and the image object given is an array then a FreeSurfer-like transform
    that places the origin at the center of the image.
    '''
    # make sure we return unchanged if no change requested
    if is_image(img) and image_type is None and spec is None and len(kwargs) == 0: return img
    elif is_image_spec(img):
        fill = kwargs.pop('fill', 0.0)
        return to_image(image_spec_to_image(img, fill=fill),
                        image_type=image_type, spec=spec, **kwargs)
    # quick cleanup of args:
    # we have a variety of things that go into spec; in order (where later overwrites earlier):
    # (1) img spec, (2) image_type map (if not an image type) (3) spec, (4) kw args
    # see if image_type is actually an image type (might be a spec/image)...
    if pimms.is_str(image_type) or isinstance(image_type, type):
        (image_type, s2) = (to_image_type(image_type), {})
    else: 
        (image_type, s2) = (None, {} if image_type is None else to_image_spec(image_type))
    if image_type is None: image_type = image_types_by_name['nifti1']
    s3 = {} if spec is None else to_image_spec(spec)
    # okay, next, parse the image argument itself:
    if is_tuple(img):
        if   len(img) == 1: (img,aff,s1) = (img[0], None, {})
        elif len(img) == 2: (img,aff,s1) = (img[0], None, img[1])
        elif len(img) == 3: (img,aff,s1) = img
        else: raise ValueError('cannot parse more than 3 elements from image tuple')
        # check that the affine wasn't given as the meta-data (e.g. (img,aff) instead of (img,mdat))
        if aff is None and s1 is not None:
            try:    (aff, s1) = (to_affine(s1, 3), {})
            except Exception: pass
    else: (aff,s1) = (None, {})
    s0 = to_image_spec(img)
    spec = pimms.merge(s0, s1, s2, s3, kwargs)
    if aff is not None: spec = pimms.assoc(spec, affine=to_affine(aff, 3))
    # okay, we create the image now:
    return image_type.create(img, meta_data=spec)
def image_copy(img, dataobj=Ellipsis, affine=Ellipsis, image_type=None):
    '''
    image_copy(image) copies the given image and returns the new object; the affine, header, and
      dataobj members are duplicated so that changes will not change the original image.
    
    The following optional arguments may be given to overwrites part of the new image; in each case,
    the default value (Ellipsis) specifies that no update should be made.
      * dataobj (default: Ellipsis) may overwrite the new image's dataobj object.
      * affine (default: Ellipsis) may overwrite the new image's affine transformation matrix.
    '''
    dataobj = np.array(img.dataobj) if dataobj is Ellipsis else np.asanyarray(dataobj)
    imspec = to_image_spec(img)
    imtype = to_image_type(img if image_type is None else image_type)
    affine = imspec['affine'] if affine is Ellipsis else affine
    imspec['affine'] = affine
    return imtype.create(dataobj, imspec)
def image_clear(img, fill=0):
    '''
    image_clear(img) yields a duplicate of the given image img but with all voxel values set to 0.
    image_clear(img, fill) sets all voxels to the given fill value.
    '''
    img = image_copy(img)
    img.dataobj[...] = fill
    return img
def is_pimage(img):
    '''
    is_pimage(img) yields True if img is an image object and it contains a persistent dataobj (i.e.,
      the dataobj is a numpy array with the writeable flag set to False); otherwise yields False.
    '''
    if   not is_image(img):                      return False
    elif not pimms.is_nparray(img.dataobj):      return False
    elif img.dataobj.flags['WRITEABLE'] is True: return False
    else:                                        return True
def is_npimage(img):
    '''
    is_npimage(img) yields True if img is an image object and its dataobj member is a numpy array--
      i.e., img is not a pointer to an array-proxy object (e.g., when the image is cached on disk);
      yields False otherwise.
    '''
    if   not is_image(img):                      return False
    elif not pimms.is_nparray(img.dataobj):      return False
    else:                                        return True
def image_interpolate(img, points, affine=None, method=None, fill=0, dtype=None, weights=None):
    '''
    image_interpolate(img, points) yields the result of interpolating the given points in the given
      image.

    Generally, the provided image (img) would be a nibabel image object, in which case, an affine
    transformation is included; if img is just an array, the affine transform is assumed to a
    FreeSurfer-like transform that places the origin at the center of the array; see to_image().

    The following options may be used:
      * affine (default: None) may specify the affine transform that aligns the vertex coordinates
        with the image (vertex-to-voxel transform). If image is an MGHImage or a Nifti1Image or
        similar, then the affine transform included in the header will be used by default if None is
        given; this parameter overwrites whatever parameter is included in the image, however.
      * method (default: None) may specify either 'linear' or 'nearest'; if None, then the
        interpolation is linear when the image data is real and nearest otherwise.
      * fill (default: 0) values filled in when a vertex falls outside of the image.
      * affine (default: None) may optionally give a final transformation that converts from vertex
        positions to native subject orientation.
      * weights (default: None) may optionally provide an image whose voxels are weights to use
        during the interpolation; these weights are in addition to trilinear weights and are
        ignored in the case of nearest interpolation unless a voxel's weight is 0. The weights,
        whether an array or an image-object, but have the same shape as the input img--any affine
        is ignored.
    '''
    points = np.asarray(points)
    if len(points.shape) == 1:
        return image_interpolate(img, np.reshape(points,[3,1]), affine=affine, method=method,
                                 fill=fill, dtype=dtype, weights=weights)[0]
    if points.shape[0] != 3: points = points.T
    if pimms.is_str(img): image = load(img)
    img = to_image(img) if affine is None else to_image(img, affine=affine)
    imspec = to_image_spec(img)
    image = img.dataobj
    # we'll use the inverse affine on the points
    affine = np.linalg.inv(imspec['affine'])
    if method is not None: method = method.lower()
    if method is None or method in ['auto', 'automatic']:
        method = 'linear' if np.issubdtype(image.dtype, np.inexact) else 'nearest'
    if dtype is None: dtype = image.dtype
    # figure out the weights...
    if weights is not None: 
        if pimms.is_str(weights): weights = load(weights).dataobj
        elif is_image(weights): weights = weights.dataobj
        else: weights = np.asanyarray(weights)
        if not np.array_equal(weights.shape, image.shape[:3]):
            raise ValueError('weights and image must have the same shape')
    # okay, these are actually pretty simple; first transform the coordinates
    xyz = np.dot(affine, np.vstack([points, np.ones([1,points.shape[1]])]))[:3]
    # remember: this might be a 4d or higher-dim image...
    res = np.full((xyz.shape[1],) + image.shape[3:], fill, dtype=dtype)
    # now find the nearest voxel centers...
    # if we are doing nearest neighbor; we're basically done already:
    image = np.asarray(image)
    imsh = np.reshape(image.shape[:3], (3,1))
    if method == 'nearest':
        ijk = np.asarray(np.round(xyz), dtype=np.int)
        ok = np.all(ijk >= 0, axis=0) & np.all(ijk < imsh, axis=0)
        if weights is not None:
            ww = weights[tuple(ijk[:,ok])]
            ok[ok] &= ~np.isclose(ww, 0)
        res[ok] = image[tuple(ijk[:,ok])]
        return res
    # otherwise, we do linear interpolation; start by finding the 8 neighboring voxels
    mins = np.floor(xyz)
    maxs = np.ceil(xyz)
    ok = np.all(mins >= 0, axis=0) & np.all(maxs < imsh, axis=0)
    (mins,maxs,xyz) = [x[:,ok] for x in (mins,maxs,xyz)]
    voxs = np.asarray([mins,
                       [mins[0], mins[1], maxs[2]],
                       [mins[0], maxs[1], mins[2]],
                       [mins[0], maxs[1], maxs[2]],
                       [maxs[0], mins[1], mins[2]],
                       [maxs[0], mins[1], maxs[2]],                           
                       [maxs[0], maxs[1], mins[2]],
                       maxs],
                      dtype=np.int)
    vals = np.asarray([image[tuple(row)] for row in voxs])
    # trilinear weights
    wgts = np.asarray([np.prod(1 - np.abs(xyz - row), axis=0) for row in voxs])
    # weight-image weights
    if weights is not None: wgts = wgts * np.asarray([weights[tuple(row)] for row in voxs])
    winv = zinv(np.sum(wgts, axis=0))
    wgts *= winv
    ok2 = ~np.isclose(winv, 0)
    ok[ok] &= ok2
    res[ok] = np.sum(wgts * vals, axis=0)[ok2]
    return res
def image_apply(image, affine, post=True):
    '''
    image_apply(im, aff) applies the given affine transform to to_image(im) and yields an equivalent
      image with the new transform in place of the old one.

    The optional third argument post (default: True) may be set to False to specify that the affine
    in the returned image should be dot(im.affine, aff) instead of dot(aff, im.affine).
    '''
    im = to_image(image)
    af = to_affine(affine, 3)
    af = np.dot(aff, im.affine) if post else np.dot(im.affine, aff)
    return to_image(image, affine=affine)
def image_reslice(image, spec, method=None, fill=0, dtype=None, weights=None, image_type=None):
    '''
    image_reslice(image, spec) yields a duplicate of the given image resliced to have the voxels
      indicated by the given image spec. Note that spec may be an image itself.

    Optional arguments that can be passed to image_interpolate() (asside from affine) are allowed
    here and are passed through.
    '''
    if image_type is None and is_image(image): image_type = to_image_type(image)
    spec = to_image_spec(spec)
    image = to_image(image)
    # we make a big mesh and interpolate at these points...
    imsh = spec['image_shape']
    (args, kw) = ([np.arange(n) for n in imsh[:3]], {'indexing': 'ij'})
    ijk = np.asarray([u.flatten() for u in np.meshgrid(*args, **kw)])
    ijk = np.dot(spec['affine'], np.vstack([ijk, np.ones([1,ijk.shape[1]])]))[:3]
    # interpolate here...
    u = image_interpolate(image, ijk, method=method, fill=fill, dtype=dtype, weights=weights)
    return to_image((np.reshape(u, imsh), spec), image_type=image_type)

    
    
